(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header{*Multiset extension of the (non-)strict order*}

theory Multiset_Extension2
imports
  Multiset2
  List2
  List_Order
  Relations 
  "~~/src/HOL/Library/Permutation"
  Transitive_Closure
  Util
begin


definition list_of :: "'a multiset \<Rightarrow> 'a list"
  where "list_of M \<equiv> SOME xs. multiset_of xs = M"

lemma multiset_of_list_of[simp]: "multiset_of (list_of M) = M"
  by (unfold list_of_def, rule someI_ex, 
    insert surj_multiset_of[unfolded surj_def, THEN spec[of _ M]], 
    auto)

lemma list_of_singleton[simp]: "list_of {#x#} = [x]"
  unfolding list_of_def
proof(rule, simp)
  fix xs
  assume id: "multiset_of xs = {#x#}"
  then obtain y ys where xs: "xs = y # ys" by (cases xs, auto)
  show "xs = [x]" 
  proof (cases ys)
    case Nil
    with id xs show ?thesis by auto
  next
    case (Cons z zs)
    from id[unfolded xs Cons]
    have "multiset_of zs + {#z#} + {#y#} = {#x#}" by auto
    from this[unfolded union_is_single] show ?thesis by auto
  qed
qed

lemma list_of_empty[simp]: "list_of {#} = []"
  unfolding list_of_def
proof(rule, simp)
  fix xs
  assume id: "multiset_of xs = {#}"
  thus "xs = []" by (cases xs, auto)
qed

lemma list_empty_inv :
  assumes "list_of A = []"
  shows "A = {#}"
proof-
  from assms and multiset_of_list_of[of A] show ?thesis by auto
qed



subsection{* Definition of the orders*}

text{* We define here A \<ge># B: the non-strict extension of the
 order "greater than" on multisets. *}

definition ns_mul_ext :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a multiset) rel" where
  "ns_mul_ext ns s \<equiv> {(A, B). (\<exists>as bs A' B'.
    A = multiset_of as + A' \<and>
    B = multiset_of bs + B' \<and>
    length as = length bs \<and>
    (\<forall>i < length bs. ((as ! i), (bs ! i)) \<in> ns) \<and>
    (\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a, b) \<in> s)))}"

text{* We define here A ># B: the strict extension of the
 order "greater than" on multisets. *}

definition s_mul_ext :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a multiset) rel"
 where "s_mul_ext ns s \<equiv> { (A,B). (\<exists> as bs A' B'.
  A = multiset_of as + A' \<and>
  B = multiset_of bs + B' \<and>
  length as = length bs \<and>
  A' \<noteq> {#} \<and>
  (\<forall>i. i < length bs \<longrightarrow> ((as ! i), (bs ! i)) \<in> ns) \<and>
  (\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s))) }"

declare ns_mul_ext_def[simp]
declare s_mul_ext_def[simp]

subsection{* Basic properties *}

lemma s_mul_ext_mono:
  assumes "ns \<subseteq> ns'" "s \<subseteq> s'"
  shows "s_mul_ext ns s \<subseteq> s_mul_ext ns' s'"
  unfolding s_mul_ext_def
  using assms by (blast 19)

lemma ns_mul_ext_mono: assumes "ns \<subseteq> ns'" "s \<subseteq> s'"
  shows "ns_mul_ext ns s \<subseteq> ns_mul_ext ns' s'" unfolding ns_mul_ext_def
  using assms by (blast 19)

lemma s_mul_ext_local_mono: assumes sub: "(set_of xs \<times> set_of ys) \<inter> ns \<subseteq> ns'" "(set_of xs \<times> set_of ys) \<inter> s \<subseteq> s'"
  and rel: "(xs,ys) \<in> s_mul_ext ns s"
  shows "(xs,ys) \<in> s_mul_ext ns' s'"
proof -
  from rel[unfolded s_mul_ext_def]
  obtain as bs A' B'
    where eq: "xs = multiset_of as + A'" 
        "ys = multiset_of bs + B'" 
         "length as = length bs" and
        ne: "A' \<noteq> {#}" and
        ns: "\<And> i. i <length bs \<Longrightarrow> (as ! i, bs ! i) \<in> ns" and
        s: "\<And> b. b \<in># B' \<Longrightarrow> (\<exists>a. a \<in># A' \<and> (a, b) \<in> s)" by blast
  show ?thesis
    unfolding s_mul_ext_def
  proof (rule, unfold split, intro exI, intro conjI allI impI, (rule eq)+, rule ne)
    fix i
    assume i: "i < length bs" with eq have "(as ! i, bs ! i) \<in> set as \<times> set bs"
      by auto
    hence "(as ! i, bs ! i) \<in> set_of xs \<times> set_of ys" unfolding eq set_of_union
      set_of_multiset_of by auto
    with ns[OF i] sub show "(as ! i, bs ! i) \<in> ns'" by auto
  next
    fix b
    assume b: "b \<in># B'"
    from s[OF this] obtain a where a: "a \<in># A'" and s: "(a,b) \<in> s" by auto
    from a b have "(a,b) \<in> set_of A' \<times> set_of B'" by auto
    hence "(a,b) \<in> set_of xs \<times> set_of ys" unfolding eq set_of_union by auto
    with a s sub show "\<exists> a. a \<in># A' \<and> (a,b) \<in> s'" by blast
  qed
qed

lemma ns_mul_ext_local_mono: assumes sub: "(set_of xs \<times> set_of ys) \<inter> ns \<subseteq> ns'" "(set_of xs \<times> set_of ys) \<inter> s \<subseteq> s'"
  and rel: "(xs,ys) \<in> ns_mul_ext ns s"
  shows "(xs,ys) \<in> ns_mul_ext ns' s'"
proof -
  from rel[unfolded ns_mul_ext_def]
  obtain as bs A' B'
    where eq: "xs = multiset_of as + A'" 
        "ys = multiset_of bs + B'" 
         "length as = length bs" and
        ns: "\<And> i. i <length bs \<Longrightarrow> (as ! i, bs ! i) \<in> ns" and
        s: "\<And> b. b \<in># B' \<Longrightarrow> (\<exists>a. a \<in># A' \<and> (a, b) \<in> s)" by blast
  show ?thesis
    unfolding ns_mul_ext_def
  proof (rule, unfold split, intro exI, intro conjI allI impI, (rule eq)+)
    fix i
    assume i: "i < length bs" with eq have "(as ! i, bs ! i) \<in> set as \<times> set bs"
      by auto
    hence "(as ! i, bs ! i) \<in> set_of xs \<times> set_of ys" unfolding eq set_of_union
      set_of_multiset_of by auto
    with ns[OF i] sub show "(as ! i, bs ! i) \<in> ns'" by auto
  next
    fix b
    assume b: "b \<in># B'"
    from s[OF this] obtain a where a: "a \<in># A'" and s: "(a,b) \<in> s" by auto
    from a b have "(a,b) \<in> set_of A' \<times> set_of B'" by auto
    hence "(a,b) \<in> set_of xs \<times> set_of ys" unfolding eq set_of_union by auto
    with a s sub show "\<exists> a. a \<in># A' \<and> (a,b) \<in> s'" by blast
  qed
qed


text{* The empty multiset is the minimal element for these orders *}

lemma ns_mul_ext_bottom: "(A,{#}) \<in> (ns_mul_ext ns s)"
by auto

lemma ns_mul_ext_bottom_uniqueness:
  assumes "({#},A) \<in> (ns_mul_ext ns s)"
  shows "A = {#}"
proof-
  from assms obtain es as E' A' where
    E_part: "{#} = multiset_of es + E'" and
    A_part: "A = multiset_of as + A'" and
    eslas: "length es = length as" and
    s_E'_A': "\<forall>a. a \<in># A' \<longrightarrow> (\<exists>e. e \<in># E' \<and> (e,a) \<in> s)" by auto
  from E_part and eslas have as_empty: "as = []" by simp
  show ?thesis proof (cases "A'")
  case empty
    with as_empty and A_part show ?thesis by simp
  next
  case (add A a)
    with s_E'_A'[rule_format, of a] obtain e where e_in_E': "e \<in># E'" by auto
    with E_part show ?thesis by simp
  qed
qed

lemma ns_mul_ext_bottom2: 
  assumes "(A,B) \<in> (ns_mul_ext ns s)"
  and "B \<noteq> {#}"
  shows "A \<noteq> {#}"
using assms proof (cases A)
case empty
  with `(A,B) \<in> (ns_mul_ext ns s)` and ns_mul_ext_bottom_uniqueness
   have "B = {#}" by fast
  with `B \<noteq> {#}` show ?thesis by simp
qed auto

lemma s_mul_ext_bottom:
  assumes "A \<noteq> {#}"
  shows "(A,{#}) \<in> (s_mul_ext ns s)"
using assms by auto

lemma s_mul_ext_bottom_strict:
  "({#},A) \<notin> (s_mul_ext ns s)"
proof-
{
  assume "({#},A) \<in> (s_mul_ext ns s)"
   then obtain as bs A' B' where
     "{#} = multiset_of as + A'" and
     "A' \<noteq> {#}" by simp
   hence "False" by simp
} thus ?thesis by simp
qed

text{* These definitions are somehow extensional. *}

lemma s_mul_ext_extensional:
  defines[simp]: "m \<equiv> multiset_of"
  assumes "m xs1 = m ys1" (is "?mxs1 = ?mys1")
  and "m xs2 = m ys2"
  and "\<forall>y y'. y \<in> set ys1 \<longrightarrow> y' \<in> set ys2 \<longrightarrow> (y, y') \<in> ns1 \<longleftrightarrow> (y, y') \<in> ns2"
  and "\<forall>y y'. y \<in> set ys1 \<longrightarrow> y' \<in> set ys2 \<longrightarrow> (y, y') \<in> s1 \<longleftrightarrow> (y, y') \<in> s2"
  shows "(m xs1, m xs2) \<in> s_mul_ext ns1 s1 = ((m ys1, m ys2) \<in> s_mul_ext ns2 s2)"
proof-
{
  assume s_xs1_xs2: "(m xs1, m xs2) \<in> s_mul_ext ns1 s1"
  then obtain as bs A' B' where
    xs1_part: "m xs1 = m as + A'" and
    xs2_part: "m xs2 = m bs + B'" and
    A'_ne: "A' \<noteq> {#}" and
    aslbs: "length as = length bs" and
    ns1_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns1" and
    s1_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s1)" by auto
  from assms(2) and xs1_part have ys1_part: "m ys1 = m as + A'" by simp
  from assms(3) and xs2_part have ys2_part: "m ys2 = m bs + B'" by simp
  {
    fix i assume i_bd: "i < length bs"
    with aslbs have i_bd': "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd'] and ys1_part and in_multiset_in_set[of "ys1" "as ! i"]
     have as_in_ys1: "as ! i \<in> set ys1" by simp
    from nth_mem_multiset_of[OF i_bd] and ys2_part and in_multiset_in_set[of "ys2" "bs ! i"]
     have bs_in_ys2: "bs ! i \<in> set ys2" by simp
    from assms(4)[rule_format, OF as_in_ys1 bs_in_ys2] and ns1_as_bs[rule_format, OF i_bd]
     have "(as ! i, bs ! i) \<in> ns2" by simp
  }
  hence ns2_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns2" by simp
  {
    fix b assume b_in_B': "b \<in># B'"
    with ys2_part and in_multiset_in_set[of ys2 b] have b_in_ys2: "b \<in> set ys2" by simp
    from b_in_B' and s1_A'_B' obtain a where
      a_in_A': "a \<in># A'" and
      s1_a_b: "(a,b) \<in> s1" by auto
    with ys1_part and in_multiset_in_set[of ys1 a] have a_in_ys1: "a \<in> set ys1" by simp
    from assms(5)[rule_format,OF a_in_ys1 b_in_ys2] and s1_a_b
     have "(a,b) \<in> s2" by simp
    with a_in_A' have "\<exists>a. a \<in># A' \<and> (a,b) \<in> s2" by fast
  }
  with ys1_part and ys2_part and aslbs and A'_ne and ns2_as_bs
   have "(m ys1, m ys2) \<in> s_mul_ext ns2 s2"
   unfolding s_mul_ext_def by auto
}
moreover
{
  assume s_ys1_ys2: "(m ys1, m ys2) \<in> s_mul_ext ns2 s2"
  then obtain as bs A' B' where
    ys1_part: "m ys1 = m as + A'" and
    ys2_part: "m ys2 = m bs + B'" and
    A'_ne: "A' \<noteq> {#}" and
    aslbs: "length as = length bs" and
    ns2_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns2" and
    s2_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s2)" by auto
  from assms(2) and ys1_part have xs1_part: "m xs1 = m as + A'" by simp
  from assms(3) and ys2_part have xs2_part: "m xs2 = m bs + B'" by simp
  {
    fix i assume i_bd: "i < length bs"
    with aslbs have i_bd': "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd'] and ys1_part and in_multiset_in_set[of "ys1" "as ! i"]
     have as_in_ys1: "as ! i \<in> set ys1" by simp
    from nth_mem_multiset_of[OF i_bd] and ys2_part and in_multiset_in_set[of "ys2" "bs ! i"]
     have bs_in_ys2: "bs ! i \<in> set ys2" by simp
    from assms(4)[rule_format, OF as_in_ys1 bs_in_ys2] and ns2_as_bs[rule_format, OF i_bd]
     have "(as ! i, bs ! i) \<in> ns1" by simp
  }
  hence ns1_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns1" by simp
  {
    fix b assume b_in_B': "b \<in># B'"
    with ys2_part and in_multiset_in_set[of ys2 b] have b_in_ys2: "b \<in> set ys2" by simp
    from b_in_B' and s2_A'_B' obtain a where
      a_in_A': "a \<in># A'" and
      s2_a_b: "(a,b) \<in> s2" by auto
    with ys1_part and in_multiset_in_set[of ys1 a] have a_in_ys1: "a \<in> set ys1" by simp
    from assms(5)[rule_format,OF a_in_ys1 b_in_ys2] and s2_a_b
     have "(a,b) \<in> s1" by simp
    with a_in_A' have "\<exists>a. a \<in># A' \<and> (a,b) \<in> s1" by fast
  }
  with xs1_part and xs2_part and aslbs and A'_ne and ns1_as_bs
   have "(m xs1, m xs2) \<in> s_mul_ext ns1 s1"
   unfolding s_mul_ext_def by auto
}
ultimately show ?thesis by blast
qed

lemma ns_mul_ext_extensional:
  defines[simp]: "m \<equiv> multiset_of"
  assumes "m xs1 = m ys1" (is "?mxs1 = ?mys1")
  and "m xs2 = m ys2"
  and "\<forall>y y'. y \<in> set ys1 \<longrightarrow> y' \<in> set ys2 \<longrightarrow> (y, y') \<in> ns1 \<longleftrightarrow> (y, y') \<in> ns2"
  and "\<forall>y y'. y \<in> set ys1 \<longrightarrow> y' \<in> set ys2 \<longrightarrow> (y, y') \<in> s1 \<longleftrightarrow> (y, y') \<in> s2"
  shows "(m xs1, m xs2) \<in> ns_mul_ext ns1 s1 = ((m ys1, m ys2) \<in> ns_mul_ext ns2 s2)"
proof-
{
  assume s_xs1_xs2: "(m xs1, m xs2) \<in> ns_mul_ext ns1 s1"
  then obtain as bs A' B' where
    xs1_part: "m xs1 = m as + A'" and
    xs2_part: "m xs2 = m bs + B'" and
    aslbs: "length as = length bs" and
    ns1_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns1" and
    s1_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s1)" by auto
  from assms(2) and xs1_part have ys1_part: "m ys1 = m as + A'" by simp
  from assms(3) and xs2_part have ys2_part: "m ys2 = m bs + B'" by simp
  {
    fix i assume i_bd: "i < length bs"
    with aslbs have i_bd': "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd'] and ys1_part and in_multiset_in_set[of "ys1" "as ! i"]
     have as_in_ys1: "as ! i \<in> set ys1" by simp
    from nth_mem_multiset_of[OF i_bd] and ys2_part and in_multiset_in_set[of "ys2" "bs ! i"]
     have bs_in_ys2: "bs ! i \<in> set ys2" by simp
    from assms(4)[rule_format, OF as_in_ys1 bs_in_ys2] and ns1_as_bs[rule_format, OF i_bd]
     have "(as ! i, bs ! i) \<in> ns2" by simp
  }
  hence ns2_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns2" by simp
  {
    fix b assume b_in_B': "b \<in># B'"
    with ys2_part and in_multiset_in_set[of ys2 b] have b_in_ys2: "b \<in> set ys2" by simp
    from b_in_B' and s1_A'_B' obtain a where
      a_in_A': "a \<in># A'" and
      s1_a_b: "(a,b) \<in> s1" by auto
    with ys1_part and in_multiset_in_set[of ys1 a] have a_in_ys1: "a \<in> set ys1" by simp
    from assms(5)[rule_format,OF a_in_ys1 b_in_ys2] and s1_a_b
     have "(a,b) \<in> s2" by simp
    with a_in_A' have "\<exists>a. a \<in># A' \<and> (a,b) \<in> s2" by fast
  }
  with ys1_part and ys2_part and aslbs and ns2_as_bs
   have "(m ys1, m ys2) \<in> ns_mul_ext ns2 s2"
   unfolding ns_mul_ext_def by auto
}
moreover
{
  assume s_ys1_ys2: "(m ys1, m ys2) \<in> ns_mul_ext ns2 s2"
  then obtain as bs A' B' where
    ys1_part: "m ys1 = m as + A'" and
    ys2_part: "m ys2 = m bs + B'" and
    aslbs: "length as = length bs" and
    ns2_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns2" and
    s2_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s2)" by auto
  from assms(2) and ys1_part have xs1_part: "m xs1 = m as + A'" by simp
  from assms(3) and ys2_part have xs2_part: "m xs2 = m bs + B'" by simp
  {
    fix i assume i_bd: "i < length bs"
    with aslbs have i_bd': "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd'] and ys1_part and in_multiset_in_set[of "ys1" "as ! i"]
     have as_in_ys1: "as ! i \<in> set ys1" by simp
    from nth_mem_multiset_of[OF i_bd] and ys2_part and in_multiset_in_set[of "ys2" "bs ! i"]
     have bs_in_ys2: "bs ! i \<in> set ys2" by simp
    from assms(4)[rule_format, OF as_in_ys1 bs_in_ys2] and ns2_as_bs[rule_format, OF i_bd]
     have "(as ! i, bs ! i) \<in> ns1" by simp
  }
  hence ns1_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns1" by simp
  {
    fix b assume b_in_B': "b \<in># B'"
    with ys2_part and in_multiset_in_set[of ys2 b] have b_in_ys2: "b \<in> set ys2" by simp
    from b_in_B' and s2_A'_B' obtain a where
      a_in_A': "a \<in># A'" and
      s2_a_b: "(a,b) \<in> s2" by auto
    with ys1_part and in_multiset_in_set[of ys1 a] have a_in_ys1: "a \<in> set ys1" by simp
    from assms(5)[rule_format,OF a_in_ys1 b_in_ys2] and s2_a_b
     have "(a,b) \<in> s1" by simp
    with a_in_A' have "\<exists>a. a \<in># A' \<and> (a,b) \<in> s1" by fast
  }
  with xs1_part and xs2_part and aslbs and ns1_as_bs
   have "(m xs1, m xs2) \<in> ns_mul_ext ns1 s1"
   unfolding s_mul_ext_def by auto
}
ultimately show ?thesis by blast
qed

text{* Obvious introduction rules. *}

lemma all_ns_ns_mul_ext:
  assumes "length as = length bs"
  and "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns"
  shows "(multiset_of as, multiset_of bs) \<in> ns_mul_ext ns s"
proof-
have "multiset_of as = multiset_of as + {#}" by simp
moreover have "multiset_of bs = multiset_of bs + {#}" by simp
moreover have "\<forall>b. b \<in># {#} \<longrightarrow> (\<exists>a. a \<in># {#} \<and> (a,b) \<in> s)" by simp
ultimately show ?thesis using assms unfolding ns_mul_ext_def by blast
qed

lemma all_s_s_mul_ext:
  assumes "A \<noteq> {#}"
  and "\<forall>b. b \<in># B \<longrightarrow> (\<exists>a. a \<in># A \<and> (a,b) \<in> s)"
  shows "(A, B) \<in> s_mul_ext ns s"
proof-
have "A = multiset_of [] + A" by simp
moreover have "B = multiset_of [] + B" by simp
moreover have "\<forall>i. i < length [] \<longrightarrow> ([] ! i, [] ! i) \<in> ns" by simp
moreover have "length [] = length []" by simp
ultimately show ?thesis using assms unfolding s_mul_ext_def by blast
qed

text{* Being stricly lesser than implies being lesser than*}

lemma s_ns_mul_ext:
  assumes "(A,B) \<in> (s_mul_ext ns s)"
  shows "(A,B) \<in> (ns_mul_ext ns s)"
using assms by auto

text{* If A is greater than B than all the Bs are dominated by an A *}

lemma ns_mul_ext_domination:
  assumes "locally_included s ns A B" (is "?lin (A + B)")
  and "(A,B) \<in> (ns_mul_ext ns s)"
  shows "\<forall>b. b \<in># B \<longrightarrow> (\<exists>a. a \<in># A \<and> (a,b) \<in> ns)"
proof-
 from `(A,B) \<in> (ns_mul_ext ns s)` obtain as bs A' B' where
   A_part: "A = multiset_of as + A'" and
   B_part: "B = multiset_of bs + B'" and
   aslbs: "length as = length bs" and   as_ns_bs: "(\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns)" and
   A'_s_B': "(\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s))" by auto
 {
   fix b assume "b \<in># multiset_of bs"
    with in_multiset_of_idx[of bs b] obtain i where
      b_in_bs: "b = bs ! i" and
      i_bd: "i < length bs" by auto
    from b_in_bs[symmetric] and as_ns_bs and i_bd have nsab: "((as ! i),b) \<in> ns" by auto
    from i_bd and aslbs[symmetric] and nth_mem_multiset_of
     have as_in_B: "(as ! i) \<in># multiset_of as" by simp
    with A_part have "(as ! i) \<in># A" by simp
    with nsab have "\<exists>a. a \<in># A \<and> (a,b) \<in> ns" by auto
 }
 hence C1: "\<forall>b. b \<in># multiset_of bs \<longrightarrow> (\<exists>a. a \<in># A \<and> (a,b) \<in> ns)" by simp
 {
   fix b assume b_in_B': "b \<in># B'"
    with A'_s_B' obtain a where
      a_in_A': "a \<in># A'" and
      sab: "(a,b) \<in> s" by auto
    with A_part have a_in_A: "a \<in># A" by simp
    from B_part and b_in_B' have b_in_B: "b \<in># B" by simp
    from sab and `?lin (A + B)`[unfolded locally_included_def, rule_format]
     and b_in_B and a_in_A have "(a,b) \<in> ns" by force
    with a_in_A have "\<exists>a. a \<in># A \<and> (a,b) \<in> ns" by force
 }
 hence C2: "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A \<and> (a,b) \<in> ns)" by simp
 from C1 and C2 and B_part show ?thesis by auto
qed

text{* The non-strict order is reflexive. *}

lemma ns_mul_ext_refl :
  assumes "locally_refl ns A" (is "?lr A")
  shows "(A,A) \<in> (ns_mul_ext ns s)"
proof -
 from `?lr A`[unfolded locally_refl_def, rule_format]
   and nth_mem_multiset_of and multiset_of_list_of[of A]
   have AnsA: "\<forall>i. i < length (list_of A) \<longrightarrow> ((list_of A ! i),(list_of A ! i)) \<in> ns" 
   by metis
 from multiset_of_list_of[symmetric] and empty_neutral[symmetric]
 have A_part: "A = multiset_of (list_of A) + {#}" by auto
 have empty: "\<forall>b. b \<in># {#} \<longrightarrow> (\<exists>a. a \<in># {#} \<and> (a,b) \<in> s)" by auto
 from AnsA and empty and A_part show ?thesis unfolding ns_mul_ext_def
  by blast
qed

text{* The orders are union-compatible *}

lemma mul_part_extend:
  assumes "A = multiset_of as + A'"
  shows "A + C = multiset_of (as @ list_of C) + A'"
using multiset_of_list_of and union_commute and union_lcomm
 and multiset_of_append and assms by metis

lemma ns_s_mul_ext_union_multiset_l:
  assumes "(A,B) \<in> (ns_mul_ext ns s)"
  and "C \<noteq> {#}"
  and "\<forall>d. d \<in># D \<longrightarrow> (\<exists>c. c \<in># C \<and> (c,d) \<in> s)"
  shows "(A + C, B + D) \<in> (s_mul_ext ns s)"
proof-
 from `(A,B) \<in> (ns_mul_ext ns s)` obtain as A' bs B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
 from A_part and union_assoc
  have A_part2: "A + C = multiset_of as + (A' + C)" by simp
 from B_part and union_assoc
  have B_part2: "B + D = multiset_of bs + (B' + D)" by simp
 from `C \<noteq> {#}` have A'C_ne: "A' + C \<noteq> {#}" by simp
 {
   fix b assume b_in: "b \<in># (B' + D)"
   have "\<exists>a. a \<in># (A' + C) \<and> (a,b) \<in> s"
   proof (cases "b \<in># B'")
   case True with s_A'_B' show ?thesis by auto
   next
   case False with b_in have "b \<in># D" by simp
     with assms(3) show ?thesis by auto
   qed
 }
 hence s_A'C_B'D: "\<forall>b. b \<in># (B' + D) \<longrightarrow> (\<exists>a. a \<in># (A' + C) \<and> (a,b) \<in> s)" by simp
 from A_part2 and B_part2 and aslbs and A'C_ne and ns_as_bs and s_A'C_B'D
  show ?thesis unfolding s_mul_ext_def by blast
qed

lemma s_mul_ext_union_compat :
  assumes "(A,B) \<in> (s_mul_ext ns s)"
  and "locally_refl ns C"      (is "?lr C") 
  shows "((A + C),(B + C)) \<in> (s_mul_ext ns s)"
proof-
  from `(A,B) \<in> (s_mul_ext ns s)` obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    A'_ne: "A' \<noteq> {#}" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
   from mul_part_extend and A_part and B_part
   have AC_part: "A + C = multiset_of (as @ list_of C) + A'"
   and BC_part: "B + C = multiset_of (bs @ list_of C) + B'" by auto
   {
     fix i assume i_bd: "i < length (list_of C)"
     from nth_mem_multiset_of[OF i_bd] and multiset_of_list_of
      have "list_of C ! i \<in># C" by simp
     with `?lr C`[unfolded locally_refl_def, rule_format]
      have "((list_of C ! i),(list_of C ! i)) \<in> ns" by simp
   }
   hence ns_c_c: "\<forall>i. i < length (list_of C) \<longrightarrow> ((list_of C ! i),(list_of C ! i)) \<in> ns"
    by simp
   from aslbs and length_append
    have "length (as @ list_of C) = length (bs @ list_of C)" by simp
   with P_as_bs_extend[OF aslbs _ ns_as_bs] and AC_part and BC_part and
     A'_ne and s_A'_B' and ns_c_c
    show ?thesis unfolding s_mul_ext_def by blast
qed

lemma ns_mul_ext_union_compat:
  assumes AnsB: "(A,B) \<in> (s_mul_ext ns s)"
  and "locally_refl ns C"      (is "?lr C") 
  shows "((A + C),(B + C)) \<in> (ns_mul_ext ns s)"
proof-
  from AnsB obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
   from mul_part_extend and A_part and B_part
   have AC_part: "A + C = multiset_of (as @ list_of C) + A'"
   and BC_part: "B + C = multiset_of (bs @ list_of C) + B'" by auto
   from aslbs and length_append
    have acslbcs: "length (as @ list_of C) = length (bs @ list_of C)"
    by simp
   {
     fix i assume i_bd: "i < length (list_of C)"
     from nth_mem_multiset_of[OF i_bd] and multiset_of_list_of
      have "list_of C ! i \<in># C" by simp
     with `?lr C`[unfolded locally_refl_def, rule_format]
      have "((list_of C ! i),(list_of C ! i)) \<in> ns" by simp
   }
   hence ns_c_c: "\<forall>i. i < length (list_of C) \<longrightarrow> ((list_of C ! i),(list_of C ! i)) \<in> ns"
    by simp
   with P_as_bs_extend[OF aslbs _ ns_as_bs] and AC_part and BC_part and
     acslbcs and s_A'_B'
    show ?thesis unfolding ns_mul_ext_def by blast
qed

text{* mapping a function preserves the order *}

lemma in_image_mset_idxE: assumes b: "b \<in># image_mset f ms"
  shows "\<exists> i. i < length (list_of ms) \<and> b = f (list_of ms ! i)"
proof -
  from b have "b \<in># image_mset f (multiset_of (list_of ms))" by simp
  from in_multiset_of_idx[OF this[unfolded multiset_of_map[symmetric]]]
  show ?thesis by auto
qed


lemma ns_mul_ext_map:
  assumes "\<And> s t. s \<in># ss \<Longrightarrow> t \<in># ts \<Longrightarrow> (s,t) \<in> S \<Longrightarrow> (f s, f t) \<in> S"
  assumes "\<And> s t. s \<in># ss \<Longrightarrow> t \<in># ts \<Longrightarrow> (s,t) \<in> NS \<Longrightarrow> (f s, f t) \<in> NS"
  and "(ss,ts) \<in> ns_mul_ext NS S"
  shows "(image_mset f ss, image_mset f ts) \<in> ns_mul_ext NS S"
proof-
  let ?f = "f"
  let ?m = "image_mset"
  from assms(3) obtain as bs A' B' where
    ss_part: "ss = multiset_of as + A'" and
    ts_part: "ts = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> NS" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> S)"
    unfolding ns_mul_ext_def by auto
  have fss_part: "?m ?f ss = multiset_of (map ?f as) + ?m ?f A'" 
    unfolding ss_part by (simp add: multiset_of_map)
  have fts_part: "?m ?f ts = multiset_of (map ?f bs) + ?m ?f B'" 
    unfolding ts_part by (simp add: multiset_of_map)
  from aslbs have faslfbs: "length (map ?f as) = length (map ?f bs)" by simp
  {
    fix i assume i_bd: "i < length (map ?f bs)"
    hence i_bd': "i < length bs" by simp
    with aslbs have i_bd'': "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd''] and ss_part 
    have as_in_ss: "as ! i \<in># ss" by simp
    from nth_mem_multiset_of[OF i_bd'] and ts_part 
    have bs_in_ts: "bs ! i \<in># ts" by simp
    from ns_as_bs[rule_format, OF i_bd'] and assms(2)[OF as_in_ss bs_in_ts]
    have "(?f (as ! i), ?f (bs ! i)) \<in> NS" by simp
    with nth_map[OF i_bd', of ?f] and nth_map[OF i_bd'', of ?f]
    have "(map ?f as ! i,  map ?f bs ! i) \<in> NS" by simp
  }
  hence ns_fas_fbs: "\<forall>i. i < length (map ?f bs) \<longrightarrow> (map ?f as ! i, map ?f bs ! i) \<in> NS" by simp
  {
    fix b assume "b \<in># ?m ?f B'"
    from in_image_mset_idxE[OF this] obtain i where
      i_bd: "i < length (list_of B')" and
      eq_b_B': "b = ?f (list_of B' ! i)" by auto
    from nth_mem_multiset_of[OF i_bd] and multiset_of_list_of
    have b_in_B': "(list_of B') ! i \<in># B'" by simp
    with ts_part
    have b_in_ts: "list_of B' ! i \<in># ts" by simp
    from s_A'_B'[rule_format, OF b_in_B'] obtain a where
      a_in_A': "a \<in># A'" and
      s_a_b: "(a,list_of B' ! i) \<in> S" by auto
    with ss_part have a_in_ss: "a \<in># ss" by simp
    from assms(1)[OF a_in_ss b_in_ts] and s_a_b
    have T: "(?f a,f (list_of B' ! i)) \<in> S" by simp
    have "\<exists>a. a \<in># ?m ?f A' \<and> (a, b) \<in> S" unfolding eq_b_B' using a_in_A' T by auto 
  }
  hence s_fA'_fB': "\<forall>b. b \<in># ?m ?f B' \<longrightarrow> (\<exists>a. a \<in># ?m ?f A' \<and> (a,b) \<in> S)" by simp
  from fss_part and fts_part and faslfbs and ns_fas_fbs and s_fA'_fB'
  show "(?m ?f ss, ?m ?f ts) \<in> (ns_mul_ext NS S)"
    unfolding ns_mul_ext_def by blast
qed

lemma s_mul_ext_map:
  assumes "\<And> s t. s \<in># ss \<Longrightarrow> t \<in># ts \<Longrightarrow> (s,t) \<in> S \<Longrightarrow> (f s, f t) \<in> S"
  assumes "\<And> s t. s \<in># ss \<Longrightarrow> t \<in># ts \<Longrightarrow> (s,t) \<in> NS \<Longrightarrow> (f s, f t) \<in> NS"
  and "(ss,ts) \<in> s_mul_ext NS S"
  shows "(image_mset f ss, image_mset f ts) \<in> s_mul_ext NS S"
proof-
  let ?f = "f"
  let ?m = "image_mset"
  from assms(3) obtain as bs A' B' where
    ss_part: "ss = multiset_of as + A'" and
    ts_part: "ts = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> NS" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> S)"
    and A': "A' \<noteq> {#}"
    unfolding s_mul_ext_def by auto
  have fss_part: "?m ?f ss = multiset_of (map ?f as) + ?m ?f A'" 
    unfolding ss_part by (simp add: multiset_of_map)
  have fts_part: "?m ?f ts = multiset_of (map ?f bs) + ?m ?f B'" 
    unfolding ts_part by (simp add: multiset_of_map)
  from aslbs have faslfbs: "length (map ?f as) = length (map ?f bs)" by simp
  {
    fix i assume i_bd: "i < length (map ?f bs)"
    hence i_bd': "i < length bs" by simp
    with aslbs have i_bd'': "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd''] and ss_part 
    have as_in_ss: "as ! i \<in># ss" by simp
    from nth_mem_multiset_of[OF i_bd'] and ts_part 
    have bs_in_ts: "bs ! i \<in># ts" by simp
    from ns_as_bs[rule_format, OF i_bd'] and assms(2)[OF as_in_ss bs_in_ts]
    have "(?f (as ! i), ?f (bs ! i)) \<in> NS" by simp
    with nth_map[OF i_bd', of ?f] and nth_map[OF i_bd'', of ?f]
    have "(map ?f as ! i,  map ?f bs ! i) \<in> NS" by simp
  }
  hence ns_fas_fbs: "\<forall>i. i < length (map ?f bs) \<longrightarrow> (map ?f as ! i, map ?f bs ! i) \<in> NS" by simp
  {
    fix b assume "b \<in># ?m ?f B'"
    from in_image_mset_idxE[OF this] obtain i where
      i_bd: "i < length (list_of B')" and
      eq_b_B': "b = ?f (list_of B' ! i)" by auto
    from nth_mem_multiset_of[OF i_bd] and multiset_of_list_of
    have b_in_B': "(list_of B') ! i \<in># B'" by simp
    with ts_part
    have b_in_ts: "list_of B' ! i \<in># ts" by simp
    from s_A'_B'[rule_format, OF b_in_B'] obtain a where
      a_in_A': "a \<in># A'" and
      s_a_b: "(a,list_of B' ! i) \<in> S" by auto
    with ss_part have a_in_ss: "a \<in># ss" by simp
    from assms(1)[OF a_in_ss b_in_ts] and s_a_b
    have T: "(?f a,f (list_of B' ! i)) \<in> S" by simp
    have "\<exists>a. a \<in># ?m ?f A' \<and> (a, b) \<in> S" unfolding eq_b_B' using a_in_A' T by auto
  }
  hence s_fA'_fB': "\<forall>b. b \<in># ?m ?f B' \<longrightarrow> (\<exists>a. a \<in># ?m ?f A' \<and> (a,b) \<in> S)" by simp
  from A' have "?m ?f A' \<noteq> {#}" by (cases A', auto)
  with fss_part and fts_part and faslfbs and ns_fas_fbs and s_fA'_fB'
  show "(?m ?f ss, ?m ?f ts) \<in> (s_mul_ext NS S)"
    unfolding s_mul_ext_def by blast
qed

lemmas mul_ext_map = s_mul_ext_map ns_mul_ext_map

definition mul_ext :: "('a \<Rightarrow> 'a \<Rightarrow> bool \<times> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool \<times> bool"
  where "mul_ext f xs ys \<equiv> let s = {(x,y). fst (f x y)}; ns = {(x,y). snd (f x y)} in ((multiset_of xs,multiset_of ys) \<in> s_mul_ext ns s, (multiset_of xs,multiset_of ys) \<in> ns_mul_ext ns s)"

lemma nstri_mul_ext_map:
  assumes "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> fst (order s t) \<Longrightarrow> fst (order (f s) (f t))"
  and "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> snd (order s t) \<Longrightarrow> snd (order (f s) (f t))"
  and "snd (mul_ext order ss ts)"
  shows "snd (mul_ext order (map f ss) (map f ts))"
  unfolding mul_ext_def Let_def snd_conv multiset_of_map
  by (rule ns_mul_ext_map, unfold in_multiset_in_set, insert assms, auto simp: mul_ext_def Let_def)

lemma stri_mul_ext_map:
  assumes "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> fst (order s t) \<Longrightarrow> fst (order (f s) (f t))"
  and "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> snd (order s t) \<Longrightarrow> snd (order (f s) (f t))"
  and "fst (mul_ext order ss ts)"
  shows "fst (mul_ext order (map f ss) (map f ts))"
  unfolding mul_ext_def Let_def fst_conv multiset_of_map
  by (rule s_mul_ext_map, unfold in_multiset_in_set, insert assms, auto simp: mul_ext_def Let_def)

text{* The non-strict order is transitive. *}

lemma ns_mul_ext_trans_ind :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and "size A = n"
  and "(A,B) \<in> (ns_mul_ext ns s)"
  and "(B,C) \<in> (ns_mul_ext ns s)"
  shows "(A,C) \<in> (ns_mul_ext ns s)"
using assms proof (induct n arbitrary: A B C)
case 0
  from `size A = 0` have "A = {#}" by simp
  with ns_mul_ext_bottom_uniqueness and `(A,B) \<in> (ns_mul_ext ns s)`
   have "B = {#}" by fast
  with ns_mul_ext_bottom_uniqueness and `(B,C) \<in> (ns_mul_ext ns s)`
   have "C = {#}" by fast
  with ns_mul_ext_bottom show ?case by simp
next
case (Suc m)
  from `(A,B) \<in> (ns_mul_ext ns s)` obtain as b1s A' B1' where
    A_part: "A = multiset_of as + A'" and
    B_part1: "B = multiset_of b1s + B1'" and
    aslb1s: "length as = length b1s" and
    ns_as_b1s: "\<forall>i. i < length b1s \<longrightarrow> ((as ! i),(b1s ! i)) \<in> ns" and
    s_A'_B1': "\<forall>b. b \<in># B1' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" unfolding ns_mul_ext_def by auto
  from `(B,C) \<in> (ns_mul_ext ns s)` obtain b2s cs B2' C' where
    B_part2: "B = multiset_of b2s + B2'" and
    C_part: "C = multiset_of cs + C'" and
    b2slcs: "length b2s = length cs" and
    ns_b2s_cs: "\<forall>i. i < length cs \<longrightarrow> ((b2s ! i),(cs ! i)) \<in> ns" and
    s_B2'_C': "\<forall>c. c \<in># C' \<longrightarrow> (\<exists>b. b \<in># B2' \<and> (b,c) \<in> s)" unfolding ns_mul_ext_def by auto
  show ?case proof (cases "\<exists>i j. i < length b1s \<and> j < length b2s \<and> b1s ! i = b2s ! j")
  case False note T = this
    have A_part': "A = multiset_of [] + A" by simp
    have C_part': "C = multiset_of [] + C" by simp
    have aslcs: "length [] = length []" by simp
    have ns_as_cs: "\<forall>i. i < length [] \<longrightarrow> (([] ! i), ([] ! i)) \<in> ns" by simp
    {
      fix c assume c_in_C: "c \<in># C"
      have "\<exists>a. a \<in># A \<and> (a,c) \<in> s"
      proof (cases "c \<in># multiset_of cs")
      case True
        from in_multiset_of_idx[OF True] obtain i where
          i_bd: "i < length cs" and
          eq_c_cs: "c = cs ! i" by fast
        with ns_b2s_cs have ns_b_c: "((b2s ! i),c) \<in> ns" by simp
        show ?thesis proof (cases "b2s ! i \<in># multiset_of b1s")
        case True
          from in_multiset_of_idx[OF True] obtain j where
            j_bd: "j < length b1s" and
            eq_b_b1s: "b2s ! i = b1s ! j" by fast
          from eq_b_b1s and T and i_bd and b2slcs and j_bd
           show ?thesis by metis
        next
        case False
          from i_bd and b2slcs and nth_mem_multiset_of
           have "b2s ! i \<in># multiset_of b2s" by metis
          with B_part2 have b_in_B: "b2s ! i \<in># B" by simp
          with False and B_part1 have "b2s ! i \<in># B1'" by simp
          with s_A'_B1' obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,(b2s ! i)) \<in> s" by force
          from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
          from `?lcr A B C`[unfolded locally_compatible_r_def,
           rule_format, OF a_in_A b_in_B c_in_C s_a_b ns_b_c]
           and a_in_A show ?thesis by fast
        qed
      next
      case False
        with c_in_C and C_part have "c \<in># C'" by simp
        with s_B2'_C' obtain b where
          b_in_B2': "b \<in># B2'" and
          s_b_c: "(b,c) \<in> s" by force
        from b_in_B2' and B_part2 have b_in_B: "b \<in># B" by simp
        {
          assume "b \<in># multiset_of b1s"
          from in_multiset_of_idx[OF this] obtain i where
            i_bd: "i < length b1s" and
            eq_b_b1s: "b = b1s ! i" by fast
          from aslb1s and i_bd and nth_mem_multiset_of[of i as] and A_part
           have as_in_A: "(as ! i) \<in># A" by simp
          from ns_as_b1s[rule_format, OF i_bd] and eq_b_b1s and as_in_A
           and `?lcl A B C`[unfolded locally_compatible_l_def, rule_format,
           OF as_in_A b_in_B c_in_C _ s_b_c]
          have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by fast
        }
        moreover
        {
          assume "b \<in># B1'"
          from s_A'_B1'[rule_format, OF this] obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,b) \<in> s" by auto
          with A_part have a_in_A: "a \<in># A" by simp
          from `?lt_s A B C`[unfolded locally_trans_def, rule_format,
           OF a_in_A b_in_B c_in_C s_a_b s_b_c] and a_in_A
           have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by auto
         }
         ultimately show ?thesis using b_in_B and B_part1 by auto
      qed
    }
    with A_part' and C_part' and aslcs and ns_as_cs
     show ?thesis unfolding ns_mul_ext_def by blast
  next
  case True note T = this
    then obtain i j where
      i_bd: "i < length b1s" and
      j_bd: "j < length b2s" and
      eq_b1s_b2s: "b1s ! i = b2s ! j" by force
    from nth_mem_multiset_of[OF i_bd] and B_part1 have b_in_B: "b1s ! i \<in># B" by simp
    from i_bd and aslb1s have i_bd2: "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd2] and A_part have a_in_A: "as ! i \<in># A" by simp
    from j_bd and b2slcs have j_bd2: "j < length cs" by simp
    from nth_mem_multiset_of[OF j_bd2] and C_part have c_in_C: "cs ! j \<in># C" by simp
    from B_part1 and nth_mem_multiset_of[OF i_bd] and diff_union_swap_multiset
     have "B - {#b1s ! i#} = (multiset_of b1s - {#b1s ! i#}) + B1'" by metis
    with remove_nth_soundness[OF i_bd]
     have B_part1': "B - {#b1s ! i#} = multiset_of (remove_nth i b1s) + B1'" by simp
    from A_part and nth_mem_multiset_of[OF i_bd2] and diff_union_swap_multiset
     have "A - {#as ! i#} = (multiset_of as - {#as ! i#}) + A'" by metis
    with remove_nth_soundness[OF i_bd2]
     have A_part': "A - {#as ! i#} = multiset_of (remove_nth i as) + A'" by simp
    from B_part2 and nth_mem_multiset_of[OF j_bd] and diff_union_swap_multiset
     have "B - {#b2s ! j#} = (multiset_of b2s - {#b2s ! j#}) + B2'" by metis
    with remove_nth_soundness[OF j_bd] and eq_b1s_b2s
     have B_part2': "B - {#b1s ! i#} = multiset_of (remove_nth j b2s) + B2'" by metis
    from C_part and nth_mem_multiset_of[OF j_bd2] and diff_union_swap_multiset
     have "C - {#cs ! j#} = (multiset_of cs - {#cs ! j#}) + C'" by metis
    with remove_nth_soundness[OF j_bd2]
     have C_part': "C - {#cs ! j#} = multiset_of (remove_nth j cs) + C'" by simp
    from remove_nth_length[OF i_bd] and remove_nth_length[OF i_bd2] and aslb1s
     have aislb1is: "length (remove_nth i as) = length (remove_nth i b1s)" by simp
    from remove_nth_length[OF j_bd] and remove_nth_length[OF j_bd2] and b2slcs
     have b2jslcjs: "length (remove_nth j b2s) = length (remove_nth j cs)" by simp      
    note ns_ais_b1is = remove_nth_P_compat[OF aslb1s[symmetric] ns_as_b1s, of i]
    note ns_b2js_cjs = remove_nth_P_compat[OF b2slcs[symmetric] ns_b2s_cs, of j]
    from A_part' and B_part1' and aislb1is and ns_ais_b1is and s_A'_B1'
     have nsAaBb: "((A - {#as ! i#}),(B - {#b1s ! i#})) \<in> (ns_mul_ext ns s)"
     unfolding ns_mul_ext_def by blast
    from B_part2' and C_part' and b2jslcjs and ns_b2js_cjs and s_B2'_C'
     have nsBbCc: "((B - {#b1s ! i#}),(C - {#cs ! j#})) \<in> (ns_mul_ext ns s)"
     unfolding ns_mul_ext_def by blast
    let ?A = "A - {#as ! i#}"
    let ?B = "B - {#b1s ! i#}"
    let ?C = "C - {#cs ! j#}"
    from lt_minus[OF `?lt_ns A B C`] have lt_ns': "?lt_ns ?A ?B ?C" by simp
    from lcl_minus[OF `?lcl A B C`] have lcl': "?lcl ?A ?B ?C" by simp
    from lcr_minus[OF `?lcr A B C`] have lcr': "?lcr ?A ?B ?C" by simp
    from lt_minus[OF `?lt_s A B C`] have lt_s': "?lt_s ?A ?B ?C" by simp
    from insert_DiffM2[OF a_in_A] and`size A = Suc m` and size_single
     and size_union[of "A - {#as ! i#}" "{#as ! i#}"]
     have "size (A - {#as ! i#}) = m" by auto
    from Suc(1)[OF lt_s' lt_ns' lcl' lcr' this nsAaBb nsBbCc]
    obtain es fs E F where
      E_part: "A - {#as ! i#} = multiset_of es + E" and
      F_part: "C - {#cs ! j#} = multiset_of fs + F" and
      eslfs: "length es = length fs" and
      ns_es_fs: "\<forall>i. i < length fs \<longrightarrow> ((es ! i),(fs ! i)) \<in> ns" and
      s_E_F: "\<forall>f. f \<in># F \<longrightarrow> (\<exists>e. e \<in># E \<and> (e,f) \<in> s)" by auto
    from E_part and insert_DiffM2[OF a_in_A] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have E_part': "A = multiset_of ((as ! i) # es) + E" by metis
    from F_part and insert_DiffM2[OF c_in_C] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have F_part': "C = multiset_of ((cs ! j) # fs) + F" by metis
    from eslfs have eslfs': "length ((as ! i) # es) = length ((cs ! j) # fs)" by simp
    {
      fix k assume k_bd: "k < length ((cs ! j) # fs)" 
      have "((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
      proof (cases k)
      case 0
        from i_bd and ns_as_b1s have ns_a_b1: "((as ! i),(b1s ! i)) \<in> ns" by simp
        from j_bd2 and ns_b2s_cs have ns_b2_c: "((b2s ! j),(cs ! j)) \<in> ns" by simp
        from `?lt_ns A B C`[unfolded locally_trans_def, rule_format,
         OF a_in_A b_in_B c_in_C ns_a_b1] and eq_b1s_b2s and ns_b2_c
         and 0
         show ?thesis by simp
      next
      case (Suc l)
        with k_bd have "l < length fs" by simp
        with ns_es_fs have "((es ! l),(fs ! l)) \<in> ns" by simp
        with Suc show ?thesis by simp
      qed
    }
    hence "\<forall>k. k < length ((cs ! j) # fs) \<longrightarrow> ((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
     by simp
    with E_part' and F_part' and eslfs' and s_E_F
     show ?thesis unfolding ns_mul_ext_def by blast
  qed
qed

lemma ns_mul_ext_trans :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and AnsB: "(A,B) \<in> (ns_mul_ext ns s)"
  and BnsC: "(B,C) \<in> (ns_mul_ext ns s)"
  shows "(A,C) \<in> (ns_mul_ext ns s)"
using assms and ns_mul_ext_trans_ind[of _ _ _ _ _ "size A"] by metis

text{* The strict order is trans. *}

lemma s_mul_ext_trans_ind :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and "size A = n"
  and "(A,B) \<in> (s_mul_ext ns s)"
  and "(B,C) \<in> (s_mul_ext ns s)"
  shows "(A,C) \<in> (s_mul_ext ns s)"
using assms proof (induct n arbitrary: A B C)
case 0
  from `size A = 0` have "A = {#}" by simp
  with s_mul_ext_bottom_strict and `(A,B) \<in> (s_mul_ext ns s)`
   show ?case by simp
next
case (Suc m)
  from Suc(6) have A_ne: "A \<noteq> {#}" by auto
  from `(A,B) \<in> (s_mul_ext ns s)` obtain as b1s A' B1' where
    A_part: "A = multiset_of as + A'" and
    B_part1: "B = multiset_of b1s + B1'" and
    aslb1s: "length as = length b1s" and
    A'_ne: "A' \<noteq> {#}" and
    ns_as_b1s: "\<forall>i. i < length b1s \<longrightarrow> ((as ! i),(b1s ! i)) \<in> ns" and
    s_A'_B1': "\<forall>b. b \<in># B1' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
  from `(B,C) \<in> (s_mul_ext ns s)` obtain b2s cs B2' C' where
    B_part2: "B = multiset_of b2s + B2'" and
    C_part: "C = multiset_of cs + C'" and
    b2slcs: "length b2s = length cs" and
    B2'_ne: "B2' \<noteq> {#}" and
    ns_b2s_cs: "\<forall>i. i < length cs \<longrightarrow> ((b2s ! i),(cs ! i)) \<in> ns" and
    s_B2'_C': "\<forall>c. c \<in># C' \<longrightarrow> (\<exists>b. b \<in># B2' \<and> (b,c) \<in> s)" by auto
  show ?case proof (cases "\<exists>i j. i < length b1s \<and> j < length b2s \<and> b1s ! i = b2s ! j")
  case False note T = this
    have A_part': "A = multiset_of [] + A" by simp
    have C_part': "C = multiset_of [] + C" by simp
    have aslcs: "length [] = length []" by simp
    have ns_as_cs: "\<forall>i. i < length [] \<longrightarrow> (([] ! i), ([] ! i)) \<in> ns" by simp
    {
      fix c assume c_in_C: "c \<in># C"
      have "\<exists>a. a \<in># A \<and> (a,c) \<in> s"
      proof (cases "c \<in># multiset_of cs")
      case True
        from in_multiset_of_idx[OF True] obtain i where
          i_bd: "i < length cs" and
          eq_c_cs: "c = cs ! i" by fast
        with ns_b2s_cs have ns_b_c: "((b2s ! i),c) \<in> ns" by simp
        show ?thesis proof (cases "b2s ! i \<in># multiset_of b1s")
        case True
          from in_multiset_of_idx[OF True] obtain j where
            j_bd: "j < length b1s" and
            eq_b_b1s: "b2s ! i = b1s ! j" by fast
          from eq_b_b1s and T and i_bd and b2slcs and j_bd
           show ?thesis by metis
        next
        case False
          from i_bd and b2slcs and nth_mem_multiset_of
           have "b2s ! i \<in># multiset_of b2s" by metis
          with B_part2 have b_in_B: "b2s ! i \<in># B" by simp
          with False and B_part1 have "b2s ! i \<in># B1'" by simp
          with s_A'_B1' obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,(b2s ! i)) \<in> s" by force
          from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
          from `?lcr A B C`[unfolded locally_compatible_r_def,
           rule_format, OF a_in_A b_in_B c_in_C s_a_b ns_b_c]
           and a_in_A show ?thesis by fast
        qed
      next
      case False
        with c_in_C and C_part have "c \<in># C'" by simp
        with s_B2'_C' obtain b where
          b_in_B2': "b \<in># B2'" and
          s_b_c: "(b,c) \<in> s" by force
        from b_in_B2' and B_part2 have b_in_B: "b \<in># B" by simp
        {
          assume "b \<in># multiset_of b1s"
          from in_multiset_of_idx[OF this] obtain i where
            i_bd: "i < length b1s" and
            eq_b_b1s: "b = b1s ! i" by fast
          from aslb1s and i_bd and nth_mem_multiset_of[of i as] and A_part
           have as_in_A: "(as ! i) \<in># A" by simp
          from ns_as_b1s[rule_format, OF i_bd] and eq_b_b1s and as_in_A
           and `?lcl A B C`[unfolded locally_compatible_l_def, rule_format,
           OF as_in_A b_in_B c_in_C _ s_b_c]
          have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by fast
        }
        moreover
        {
          assume "b \<in># B1'"
          from s_A'_B1'[rule_format, OF this] obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,b) \<in> s" by auto
          with A_part have a_in_A: "a \<in># A" by simp
          from `?lt_s A B C`[unfolded locally_trans_def, rule_format,
           OF a_in_A b_in_B c_in_C s_a_b s_b_c] and a_in_A
           have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by auto
         }
         ultimately show ?thesis using b_in_B and B_part1 by auto
      qed
    }
    with A_part' and C_part' and aslcs and A_ne and ns_as_cs
     show ?thesis unfolding s_mul_ext_def by blast
  next
  case True note T = this
    then obtain i j where
      i_bd: "i < length b1s" and
      j_bd: "j < length b2s" and
      eq_b1s_b2s: "b1s ! i = b2s ! j" by force
    from nth_mem_multiset_of[OF i_bd] and B_part1 have b_in_B: "b1s ! i \<in># B" by simp
    from i_bd and aslb1s have i_bd2: "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd2] and A_part have a_in_A: "as ! i \<in># A" by simp
    from j_bd and b2slcs have j_bd2: "j < length cs" by simp
    from nth_mem_multiset_of[OF j_bd2] and C_part have c_in_C: "cs ! j \<in># C" by simp
    from B_part1 and nth_mem_multiset_of[OF i_bd] and diff_union_swap_multiset
     have "B - {#b1s ! i#} = (multiset_of b1s - {#b1s ! i#}) + B1'" by metis
    with remove_nth_soundness[OF i_bd]
     have B_part1': "B - {#b1s ! i#} = multiset_of (remove_nth i b1s) + B1'" by simp
    from A_part and nth_mem_multiset_of[OF i_bd2] and diff_union_swap_multiset
     have "A - {#as ! i#} = (multiset_of as - {#as ! i#}) + A'" by metis
    with remove_nth_soundness[OF i_bd2]
     have A_part': "A - {#as ! i#} = multiset_of (remove_nth i as) + A'" by simp
    from B_part2 and nth_mem_multiset_of[OF j_bd] and diff_union_swap_multiset
     have "B - {#b2s ! j#} = (multiset_of b2s - {#b2s ! j#}) + B2'" by metis
    with remove_nth_soundness[OF j_bd] and eq_b1s_b2s
     have B_part2': "B - {#b1s ! i#} = multiset_of (remove_nth j b2s) + B2'" by metis
    from C_part and nth_mem_multiset_of[OF j_bd2] and diff_union_swap_multiset
     have "C - {#cs ! j#} = (multiset_of cs - {#cs ! j#}) + C'" by metis
    with remove_nth_soundness[OF j_bd2]
     have C_part': "C - {#cs ! j#} = multiset_of (remove_nth j cs) + C'" by simp
    from remove_nth_length[OF i_bd] and remove_nth_length[OF i_bd2] and aslb1s
     have aislb1is: "length (remove_nth i as) = length (remove_nth i b1s)" by simp
    from remove_nth_length[OF j_bd] and remove_nth_length[OF j_bd2] and b2slcs
     have b2jslcjs: "length (remove_nth j b2s) = length (remove_nth j cs)" by simp      
    note ns_ais_b1is = remove_nth_P_compat[OF aslb1s[symmetric] ns_as_b1s, of i]
    note ns_b2js_cjs = remove_nth_P_compat[OF b2slcs[symmetric] ns_b2s_cs, of j]
    from A_part' and B_part1' and aislb1is and A'_ne and ns_ais_b1is and s_A'_B1'
     have sAaBb: "((A - {#as ! i#}),(B - {#b1s ! i#})) \<in> (s_mul_ext ns s)"
     unfolding s_mul_ext_def by blast
    from B_part2' and C_part' and b2jslcjs and B2'_ne and ns_b2js_cjs and s_B2'_C'
     have sBbCc: "((B - {#b1s ! i#}),(C - {#cs ! j#})) \<in> (s_mul_ext ns s)"
     unfolding s_mul_ext_def by blast
    let ?A = "A - {#as ! i#}"
    let ?B = "B - {#b1s ! i#}"
    let ?C = "C - {#cs ! j#}"
    from lt_minus[OF `?lt_ns A B C`] have lt_ns': "?lt_ns ?A ?B ?C" by simp
    from lcl_minus[OF `?lcl A B C`] have lcl': "?lcl ?A ?B ?C" by simp
    from lcr_minus[OF `?lcr A B C`] have lcr': "?lcr ?A ?B ?C" by simp
    from lt_minus[OF `?lt_s A B C`] have lt_s': "?lt_s ?A ?B ?C" by simp
    from insert_DiffM2[OF a_in_A] and`size A = Suc m` and size_single
     and size_union[of "A - {#as ! i#}" "{#as ! i#}"]
     have "size (A - {#as ! i#}) = m" by auto
    from Suc(1)[OF lt_s' lt_ns' lcl' lcr' this sAaBb sBbCc]
    obtain es fs E F where
      E_part: "A - {#as ! i#} = multiset_of es + E" and
      F_part: "C - {#cs ! j#} = multiset_of fs + F" and
      eslfs: "length es = length fs" and
      E_ne: "E \<noteq> {#}" and
      ns_es_fs: "\<forall>i. i < length fs \<longrightarrow> ((es ! i),(fs ! i)) \<in> ns" and
      s_E_F: "\<forall>f. f \<in># F \<longrightarrow> (\<exists>e. e \<in># E \<and> (e,f) \<in> s)" by auto
    from E_part and insert_DiffM2[OF a_in_A] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have E_part': "A = multiset_of ((as ! i) # es) + E" by metis
    from F_part and insert_DiffM2[OF c_in_C] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have F_part': "C = multiset_of ((cs ! j) # fs) + F" by metis
    from eslfs have eslfs': "length ((as ! i) # es) = length ((cs ! j) # fs)" by simp
    {
      fix k assume k_bd: "k < length ((cs ! j) # fs)" 
      have "((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
      proof (cases k)
      case 0
        from i_bd and ns_as_b1s have ns_a_b1: "((as ! i),(b1s ! i)) \<in> ns" by simp
        from j_bd2 and ns_b2s_cs have ns_b2_c: "((b2s ! j),(cs ! j)) \<in> ns" by simp
        from `?lt_ns A B C`[unfolded locally_trans_def, rule_format,
         OF a_in_A b_in_B c_in_C ns_a_b1] and eq_b1s_b2s and ns_b2_c
         and 0
         show ?thesis by simp
      next
      case (Suc l)
        with k_bd have "l < length fs" by simp
        with ns_es_fs have "((es ! l),(fs ! l)) \<in> ns" by simp
        with Suc show ?thesis by simp
      qed
    }
    hence "\<forall>k. k < length ((cs ! j) # fs) \<longrightarrow> ((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
     by simp
    with E_part' and F_part' and eslfs' and E_ne and s_E_F
     show ?thesis unfolding s_mul_ext_def by blast
  qed
qed

lemma s_mul_ext_trans :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and AnsB: "(A,B) \<in> (s_mul_ext ns s)"
  and BnsC: "(B,C) \<in> (s_mul_ext ns s)"
  shows "(A,C) \<in> (s_mul_ext ns s)"
using assms and s_mul_ext_trans_ind[of _ _ _ _ _ "size A"] by metis

text{* The strict order is compatible on the left with the non strict one *}

lemma s_ns_mul_ext_trans_ind :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and "size A = n"
  and "(A,B) \<in> (s_mul_ext ns s)"
  and "(B,C) \<in> (ns_mul_ext ns s)"
  shows "(A,C) \<in> (s_mul_ext ns s)"
using assms proof (induct n arbitrary: A B C)
case 0
  from `size A = 0` have "A = {#}" by simp
  with s_mul_ext_bottom_strict and `(A,B) \<in> (s_mul_ext ns s)`
   show ?case by simp
next
case (Suc m)
  from Suc(6) have A_ne: "A \<noteq> {#}" by auto
  from `(A,B) \<in> (s_mul_ext ns s)` obtain as b1s A' B1' where
    A_part: "A = multiset_of as + A'" and
    B_part1: "B = multiset_of b1s + B1'" and
    aslb1s: "length as = length b1s" and
    A'_ne: "A' \<noteq> {#}" and
    ns_as_b1s: "\<forall>i. i < length b1s \<longrightarrow> ((as ! i),(b1s ! i)) \<in> ns" and
    s_A'_B1': "\<forall>b. b \<in># B1' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
  from `(B,C) \<in> (ns_mul_ext ns s)` obtain b2s cs B2' C' where
    B_part2: "B = multiset_of b2s + B2'" and
    C_part: "C = multiset_of cs + C'" and
    b2slcs: "length b2s = length cs" and
    ns_b2s_cs: "\<forall>i. i < length cs \<longrightarrow> ((b2s ! i),(cs ! i)) \<in> ns" and
    s_B2'_C': "\<forall>c. c \<in># C' \<longrightarrow> (\<exists>b. b \<in># B2' \<and> (b,c) \<in> s)" by auto
  show ?case proof (cases "\<exists>i j. i < length b1s \<and> j < length b2s \<and> b1s ! i = b2s ! j")
  case False note T = this
    have A_part': "A = multiset_of [] + A" by simp
    have C_part': "C = multiset_of [] + C" by simp
    have aslcs: "length [] = length []" by simp
    have ns_as_cs: "\<forall>i. i < length [] \<longrightarrow> (([] ! i), ([] ! i)) \<in> ns" by simp
    {
      fix c assume c_in_C: "c \<in># C"
      have "\<exists>a. a \<in># A \<and> (a,c) \<in> s"
      proof (cases "c \<in># multiset_of cs")
      case True
        from in_multiset_of_idx[OF True] obtain i where
          i_bd: "i < length cs" and
          eq_c_cs: "c = cs ! i" by fast
        with ns_b2s_cs have ns_b_c: "((b2s ! i),c) \<in> ns" by simp
        show ?thesis proof (cases "b2s ! i \<in># multiset_of b1s")
        case True
          from in_multiset_of_idx[OF True] obtain j where
            j_bd: "j < length b1s" and
            eq_b_b1s: "b2s ! i = b1s ! j" by fast
          from eq_b_b1s and T and i_bd and b2slcs and j_bd
           show ?thesis by metis
        next
        case False
          from i_bd and b2slcs and nth_mem_multiset_of
           have "b2s ! i \<in># multiset_of b2s" by metis
          with B_part2 have b_in_B: "b2s ! i \<in># B" by simp
          with False and B_part1 have "b2s ! i \<in># B1'" by simp
          with s_A'_B1' obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,(b2s ! i)) \<in> s" by force
          from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
          from `?lcr A B C`[unfolded locally_compatible_r_def,
           rule_format, OF a_in_A b_in_B c_in_C s_a_b ns_b_c]
           and a_in_A show ?thesis by fast
        qed
      next
      case False
        with c_in_C and C_part have "c \<in># C'" by simp
        with s_B2'_C' obtain b where
          b_in_B2': "b \<in># B2'" and
          s_b_c: "(b,c) \<in> s" by force
        from b_in_B2' and B_part2 have b_in_B: "b \<in># B" by simp
        {
          assume "b \<in># multiset_of b1s"
          from in_multiset_of_idx[OF this] obtain i where
            i_bd: "i < length b1s" and
            eq_b_b1s: "b = b1s ! i" by fast
          from aslb1s and i_bd and nth_mem_multiset_of[of i as] and A_part
           have as_in_A: "(as ! i) \<in># A" by simp
          from ns_as_b1s[rule_format, OF i_bd] and eq_b_b1s and as_in_A
           and `?lcl A B C`[unfolded locally_compatible_l_def, rule_format,
           OF as_in_A b_in_B c_in_C _ s_b_c]
          have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by fast
        }
        moreover
        {
          assume "b \<in># B1'"
          from s_A'_B1'[rule_format, OF this] obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,b) \<in> s" by auto
          with A_part have a_in_A: "a \<in># A" by simp
          from `?lt_s A B C`[unfolded locally_trans_def, rule_format,
           OF a_in_A b_in_B c_in_C s_a_b s_b_c] and a_in_A
           have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by auto
         }
         ultimately show ?thesis using b_in_B and B_part1 by auto
      qed
    }
    with A_part' and C_part' and aslcs and A_ne and ns_as_cs
     show ?thesis unfolding s_mul_ext_def by blast
  next
  case True note T = this
    then obtain i j where
      i_bd: "i < length b1s" and
      j_bd: "j < length b2s" and
      eq_b1s_b2s: "b1s ! i = b2s ! j" by force
    from nth_mem_multiset_of[OF i_bd] and B_part1 have b_in_B: "b1s ! i \<in># B" by simp
    from i_bd and aslb1s have i_bd2: "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd2] and A_part have a_in_A: "as ! i \<in># A" by simp
    from j_bd and b2slcs have j_bd2: "j < length cs" by simp
    from nth_mem_multiset_of[OF j_bd2] and C_part have c_in_C: "cs ! j \<in># C" by simp
    from B_part1 and nth_mem_multiset_of[OF i_bd] and diff_union_swap_multiset
     have "B - {#b1s ! i#} = (multiset_of b1s - {#b1s ! i#}) + B1'" by metis
    with remove_nth_soundness[OF i_bd]
     have B_part1': "B - {#b1s ! i#} = multiset_of (remove_nth i b1s) + B1'" by simp
    from A_part and nth_mem_multiset_of[OF i_bd2] and diff_union_swap_multiset
     have "A - {#as ! i#} = (multiset_of as - {#as ! i#}) + A'" by metis
    with remove_nth_soundness[OF i_bd2]
     have A_part': "A - {#as ! i#} = multiset_of (remove_nth i as) + A'" by simp
    from B_part2 and nth_mem_multiset_of[OF j_bd] and diff_union_swap_multiset
     have "B - {#b2s ! j#} = (multiset_of b2s - {#b2s ! j#}) + B2'" by metis
    with remove_nth_soundness[OF j_bd] and eq_b1s_b2s
     have B_part2': "B - {#b1s ! i#} = multiset_of (remove_nth j b2s) + B2'" by metis
    from C_part and nth_mem_multiset_of[OF j_bd2] and diff_union_swap_multiset
     have "C - {#cs ! j#} = (multiset_of cs - {#cs ! j#}) + C'" by metis
    with remove_nth_soundness[OF j_bd2]
     have C_part': "C - {#cs ! j#} = multiset_of (remove_nth j cs) + C'" by simp
    from remove_nth_length[OF i_bd] and remove_nth_length[OF i_bd2] and aslb1s
     have aislb1is: "length (remove_nth i as) = length (remove_nth i b1s)" by simp
    from remove_nth_length[OF j_bd] and remove_nth_length[OF j_bd2] and b2slcs
     have b2jslcjs: "length (remove_nth j b2s) = length (remove_nth j cs)" by simp      
    note ns_ais_b1is = remove_nth_P_compat[OF aslb1s[symmetric] ns_as_b1s, of i]
    note ns_b2js_cjs = remove_nth_P_compat[OF b2slcs[symmetric] ns_b2s_cs, of j]
    from A_part' and B_part1' and aislb1is and A'_ne and ns_ais_b1is and s_A'_B1'
     have sAaBb: "((A - {#as ! i#}),(B - {#b1s ! i#})) \<in> (s_mul_ext ns s)"
     unfolding s_mul_ext_def by blast
    from B_part2' and C_part' and b2jslcjs and ns_b2js_cjs and s_B2'_C'
     have nsBbCc: "((B - {#b1s ! i#}),(C - {#cs ! j#})) \<in> (ns_mul_ext ns s)"
     unfolding ns_mul_ext_def by blast
    let ?A = "A - {#as ! i#}"
    let ?B = "B - {#b1s ! i#}"
    let ?C = "C - {#cs ! j#}"
    from lt_minus[OF `?lt_ns A B C`] have lt_ns': "?lt_ns ?A ?B ?C" by simp
    from lcl_minus[OF `?lcl A B C`] have lcl': "?lcl ?A ?B ?C" by simp
    from lcr_minus[OF `?lcr A B C`] have lcr': "?lcr ?A ?B ?C" by simp
    from lt_minus[OF `?lt_s A B C`] have lt_s': "?lt_s ?A ?B ?C" by simp
    from insert_DiffM2[OF a_in_A] and`size A = Suc m` and size_single
     and size_union[of "A - {#as ! i#}" "{#as ! i#}"]
     have "size (A - {#as ! i#}) = m" by auto
    from Suc(1)[OF lt_s' lt_ns' lcl' lcr' this sAaBb nsBbCc]
    obtain es fs E F where
      E_part: "A - {#as ! i#} = multiset_of es + E" and
      F_part: "C - {#cs ! j#} = multiset_of fs + F" and
      eslfs: "length es = length fs" and
      E_ne: "E \<noteq> {#}" and
      ns_es_fs: "\<forall>i. i < length fs \<longrightarrow> ((es ! i),(fs ! i)) \<in> ns" and
      s_E_F: "\<forall>f. f \<in># F \<longrightarrow> (\<exists>e. e \<in># E \<and> (e,f) \<in> s)" by auto
    from E_part and insert_DiffM2[OF a_in_A] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have E_part': "A = multiset_of ((as ! i) # es) + E" by metis
    from F_part and insert_DiffM2[OF c_in_C] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have F_part': "C = multiset_of ((cs ! j) # fs) + F" by metis
    from eslfs have eslfs': "length ((as ! i) # es) = length ((cs ! j) # fs)" by simp
    {
      fix k assume k_bd: "k < length ((cs ! j) # fs)" 
      have "((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
      proof (cases k)
      case 0
        from i_bd and ns_as_b1s have ns_a_b1: "((as ! i),(b1s ! i)) \<in> ns" by simp
        from j_bd2 and ns_b2s_cs have ns_b2_c: "((b2s ! j),(cs ! j)) \<in> ns" by simp
        from `?lt_ns A B C`[unfolded locally_trans_def, rule_format,
         OF a_in_A b_in_B c_in_C ns_a_b1] and eq_b1s_b2s and ns_b2_c
         and 0
         show ?thesis by simp
      next
      case (Suc l)
        with k_bd have "l < length fs" by simp
        with ns_es_fs have "((es ! l),(fs ! l)) \<in> ns" by simp
        with Suc show ?thesis by simp
      qed
    }
    hence "\<forall>k. k < length ((cs ! j) # fs) \<longrightarrow> ((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
     by simp
    with E_part' and F_part' and eslfs' and E_ne and s_E_F
     show ?thesis unfolding s_mul_ext_def by blast
  qed
qed

lemma s_ns_mul_ext_trans :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and AnsB: "(A,B) \<in> (s_mul_ext ns s)"
  and BnsC: "(B,C) \<in> (ns_mul_ext ns s)"
  shows "(A,C) \<in> (s_mul_ext ns s)"
using assms and s_ns_mul_ext_trans_ind[of _ _ _ _ _ "size A"] by metis

text{* The strict order is compatible on the right with the non-strict one. *}

lemma ns_s_mul_ext_trans_ind :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and "size A = n"
  and "(A,B) \<in> (ns_mul_ext ns s)"
  and "(B,C) \<in> (s_mul_ext ns s)"
  shows "(A,C) \<in> (s_mul_ext ns s)"
using assms proof (induct n arbitrary: A B C)
case 0
  from `size A = 0` have "A = {#}" by simp
  with ns_mul_ext_bottom_uniqueness and `(A,B) \<in> (ns_mul_ext ns s)`
   have "B = {#}" by fast
  with s_mul_ext_bottom_strict and `(B,C) \<in> (s_mul_ext ns s)`
   show ?case by simp
next
case (Suc m)
  from Suc(6) have A_ne: "A \<noteq> {#}" by auto
  from `(A,B) \<in> (ns_mul_ext ns s)` obtain as b1s A' B1' where
    A_part: "A = multiset_of as + A'" and
    B_part1: "B = multiset_of b1s + B1'" and
    aslb1s: "length as = length b1s" and
    ns_as_b1s: "\<forall>i. i < length b1s \<longrightarrow> ((as ! i),(b1s ! i)) \<in> ns" and
    s_A'_B1': "\<forall>b. b \<in># B1' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
  from `(B,C) \<in> (s_mul_ext ns s)` obtain b2s cs B2' C' where
    B_part2: "B = multiset_of b2s + B2'" and
    C_part: "C = multiset_of cs + C'" and
    b2slcs: "length b2s = length cs" and
    B2'_ne: "B2' \<noteq> {#}" and
    ns_b2s_cs: "\<forall>i. i < length cs \<longrightarrow> ((b2s ! i),(cs ! i)) \<in> ns" and
    s_B2'_C': "\<forall>c. c \<in># C' \<longrightarrow> (\<exists>b. b \<in># B2' \<and> (b,c) \<in> s)" by auto
  show ?case proof (cases "\<exists>i j. i < length b1s \<and> j < length b2s \<and> b1s ! i = b2s ! j")
  case False note T = this
    have A_part': "A = multiset_of [] + A" by simp
    have C_part': "C = multiset_of [] + C" by simp
    have aslcs: "length [] = length []" by simp
    have ns_as_cs: "\<forall>i. i < length [] \<longrightarrow> (([] ! i), ([] ! i)) \<in> ns" by simp
    {
      fix c assume c_in_C: "c \<in># C"
      have "\<exists>a. a \<in># A \<and> (a,c) \<in> s"
      proof (cases "c \<in># multiset_of cs")
      case True
        from in_multiset_of_idx[OF True] obtain i where
          i_bd: "i < length cs" and
          eq_c_cs: "c = cs ! i" by fast
        with ns_b2s_cs have ns_b_c: "((b2s ! i),c) \<in> ns" by simp
        show ?thesis proof (cases "b2s ! i \<in># multiset_of b1s")
        case True
          from in_multiset_of_idx[OF True] obtain j where
            j_bd: "j < length b1s" and
            eq_b_b1s: "b2s ! i = b1s ! j" by fast
          from eq_b_b1s and T and i_bd and b2slcs and j_bd
           show ?thesis by metis
        next
        case False
          from i_bd and b2slcs and nth_mem_multiset_of
           have "b2s ! i \<in># multiset_of b2s" by metis
          with B_part2 have b_in_B: "b2s ! i \<in># B" by simp
          with False and B_part1 have "b2s ! i \<in># B1'" by simp
          with s_A'_B1' obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,(b2s ! i)) \<in> s" by force
          from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
          from `?lcr A B C`[unfolded locally_compatible_r_def,
           rule_format, OF a_in_A b_in_B c_in_C s_a_b ns_b_c]
           and a_in_A show ?thesis by fast
        qed
      next
      case False
        with c_in_C and C_part have "c \<in># C'" by simp
        with s_B2'_C' obtain b where
          b_in_B2': "b \<in># B2'" and
          s_b_c: "(b,c) \<in> s" by force
        from b_in_B2' and B_part2 have b_in_B: "b \<in># B" by simp
        {
          assume "b \<in># multiset_of b1s"
          from in_multiset_of_idx[OF this] obtain i where
            i_bd: "i < length b1s" and
            eq_b_b1s: "b = b1s ! i" by fast
          from aslb1s and i_bd and nth_mem_multiset_of[of i as] and A_part
           have as_in_A: "(as ! i) \<in># A" by simp
          from ns_as_b1s[rule_format, OF i_bd] and eq_b_b1s and as_in_A
           and `?lcl A B C`[unfolded locally_compatible_l_def, rule_format,
           OF as_in_A b_in_B c_in_C _ s_b_c]
          have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by fast
        }
        moreover
        {
          assume "b \<in># B1'"
          from s_A'_B1'[rule_format, OF this] obtain a where
            a_in_A': "a \<in># A'" and
            s_a_b: "(a,b) \<in> s" by auto
          with A_part have a_in_A: "a \<in># A" by simp
          from `?lt_s A B C`[unfolded locally_trans_def, rule_format,
           OF a_in_A b_in_B c_in_C s_a_b s_b_c] and a_in_A
           have "\<exists>a. a \<in># A \<and> (a,c) \<in> s" by auto
         }
         ultimately show ?thesis using b_in_B and B_part1 by auto
      qed
    }
    with A_part' and C_part' and aslcs and A_ne and ns_as_cs
     show ?thesis unfolding s_mul_ext_def by blast
  next
  case True note T = this
    then obtain i j where
      i_bd: "i < length b1s" and
      j_bd: "j < length b2s" and
      eq_b1s_b2s: "b1s ! i = b2s ! j" by force
    from nth_mem_multiset_of[OF i_bd] and B_part1 have b_in_B: "b1s ! i \<in># B" by simp
    from i_bd and aslb1s have i_bd2: "i < length as" by simp
    from nth_mem_multiset_of[OF i_bd2] and A_part have a_in_A: "as ! i \<in># A" by simp
    from j_bd and b2slcs have j_bd2: "j < length cs" by simp
    from nth_mem_multiset_of[OF j_bd2] and C_part have c_in_C: "cs ! j \<in># C" by simp
    from B_part1 and nth_mem_multiset_of[OF i_bd] and diff_union_swap_multiset
     have "B - {#b1s ! i#} = (multiset_of b1s - {#b1s ! i#}) + B1'" by metis
    with remove_nth_soundness[OF i_bd]
     have B_part1': "B - {#b1s ! i#} = multiset_of (remove_nth i b1s) + B1'" by simp
    from A_part and nth_mem_multiset_of[OF i_bd2] and diff_union_swap_multiset
     have "A - {#as ! i#} = (multiset_of as - {#as ! i#}) + A'" by metis
    with remove_nth_soundness[OF i_bd2]
     have A_part': "A - {#as ! i#} = multiset_of (remove_nth i as) + A'" by simp
    from B_part2 and nth_mem_multiset_of[OF j_bd] and diff_union_swap_multiset
     have "B - {#b2s ! j#} = (multiset_of b2s - {#b2s ! j#}) + B2'" by metis
    with remove_nth_soundness[OF j_bd] and eq_b1s_b2s
     have B_part2': "B - {#b1s ! i#} = multiset_of (remove_nth j b2s) + B2'" by metis
    from C_part and nth_mem_multiset_of[OF j_bd2] and diff_union_swap_multiset
     have "C - {#cs ! j#} = (multiset_of cs - {#cs ! j#}) + C'" by metis
    with remove_nth_soundness[OF j_bd2]
     have C_part': "C - {#cs ! j#} = multiset_of (remove_nth j cs) + C'" by simp
    from remove_nth_length[OF i_bd] and remove_nth_length[OF i_bd2] and aslb1s
     have aislb1is: "length (remove_nth i as) = length (remove_nth i b1s)" by simp
    from remove_nth_length[OF j_bd] and remove_nth_length[OF j_bd2] and b2slcs
     have b2jslcjs: "length (remove_nth j b2s) = length (remove_nth j cs)" by simp      
    note ns_ais_b1is = remove_nth_P_compat[OF aslb1s[symmetric] ns_as_b1s, of i]
    note ns_b2js_cjs = remove_nth_P_compat[OF b2slcs[symmetric] ns_b2s_cs, of j]
    from A_part' and B_part1' and aislb1is and ns_ais_b1is and s_A'_B1'
     have nsAaBb: "((A - {#as ! i#}),(B - {#b1s ! i#})) \<in> (ns_mul_ext ns s)"
     unfolding ns_mul_ext_def by blast
    from B_part2' and C_part' and b2jslcjs and B2'_ne and ns_b2js_cjs and s_B2'_C'
     have sBbCc: "((B - {#b1s ! i#}),(C - {#cs ! j#})) \<in> (s_mul_ext ns s)"
     unfolding s_mul_ext_def by blast
    let ?A = "A - {#as ! i#}"
    let ?B = "B - {#b1s ! i#}"
    let ?C = "C - {#cs ! j#}"
    from lt_minus[OF `?lt_ns A B C`] have lt_ns': "?lt_ns ?A ?B ?C" by simp
    from lcl_minus[OF `?lcl A B C`] have lcl': "?lcl ?A ?B ?C" by simp
    from lcr_minus[OF `?lcr A B C`] have lcr': "?lcr ?A ?B ?C" by simp
    from lt_minus[OF `?lt_s A B C`] have lt_s': "?lt_s ?A ?B ?C" by simp
    from insert_DiffM2[OF a_in_A] and`size A = Suc m` and size_single
     and size_union[of "A - {#as ! i#}" "{#as ! i#}"]
     have "size (A - {#as ! i#}) = m" by auto
    from Suc(1)[OF lt_s' lt_ns' lcl' lcr' this nsAaBb sBbCc]
    obtain es fs E F where
      E_part: "A - {#as ! i#} = multiset_of es + E" and
      F_part: "C - {#cs ! j#} = multiset_of fs + F" and
      eslfs: "length es = length fs" and
      E_ne: "E \<noteq> {#}" and
      ns_es_fs: "\<forall>i. i < length fs \<longrightarrow> ((es ! i),(fs ! i)) \<in> ns" and
      s_E_F: "\<forall>f. f \<in># F \<longrightarrow> (\<exists>e. e \<in># E \<and> (e,f) \<in> s)" by auto
    from E_part and insert_DiffM2[OF a_in_A] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have E_part': "A = multiset_of ((as ! i) # es) + E" by metis
    from F_part and insert_DiffM2[OF c_in_C] and union_commute
     and union_lcomm and multiset_of.simps(2)
     have F_part': "C = multiset_of ((cs ! j) # fs) + F" by metis
    from eslfs have eslfs': "length ((as ! i) # es) = length ((cs ! j) # fs)" by simp
    {
      fix k assume k_bd: "k < length ((cs ! j) # fs)" 
      have "((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
      proof (cases k)
      case 0
        from i_bd and ns_as_b1s have ns_a_b1: "((as ! i),(b1s ! i)) \<in> ns" by simp
        from j_bd2 and ns_b2s_cs have ns_b2_c: "((b2s ! j),(cs ! j)) \<in> ns" by simp
        from `?lt_ns A B C`[unfolded locally_trans_def, rule_format,
         OF a_in_A b_in_B c_in_C ns_a_b1] and eq_b1s_b2s and ns_b2_c
         and 0
         show ?thesis by simp
      next
      case (Suc l)
        with k_bd have "l < length fs" by simp
        with ns_es_fs have "((es ! l),(fs ! l)) \<in> ns" by simp
        with Suc show ?thesis by simp
      qed
    }
    hence "\<forall>k. k < length ((cs ! j) # fs) \<longrightarrow> ((((as ! i) # es) ! k),(((cs ! j) # fs) ! k)) \<in> ns"
     by simp
    with E_part' and F_part' and eslfs' and E_ne and s_E_F
     show ?thesis unfolding s_mul_ext_def by blast
  qed
qed

lemma ns_s_mul_ext_trans :
  assumes "locally_trans s A B C"  (is "?lt_s A B C")
  and "locally_trans ns A B C"     (is "?lt_ns A B C")
  and "locally_compatible_l ns s A B C" (is "?lcl A B C")
  and "locally_compatible_r ns s A B C" (is "?lcr A B C")
  and AnsB: "(A,B) \<in> (ns_mul_ext ns s)"
  and BnsC: "(B,C) \<in> (s_mul_ext ns s)"
  shows "(A,C) \<in> (s_mul_ext ns s)"
using assms and ns_s_mul_ext_trans_ind[of _ _ _ _ _ "size A"] by metis

text{* To prove that s_mul_ext is well-founded, we first define
 step_s_mul_ext whose trans closure IS s_mul_ext and which can
 (quite painfully but rather easily) be proved SN. *}

definition step_s_mul_ext :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a multiset rel" where
  "step_s_mul_ext ns s \<equiv> { (P,Q). \<exists>a as ps qs.
            P = multiset_of ps + {#a#} \<and>
            Q = multiset_of qs + multiset_of as \<and>
            length ps = length qs \<and>
            (\<forall>i. i < length qs \<longrightarrow> ((ps ! i),(qs ! i)) \<in> ns) \<and>
            (\<forall>i. i < length as \<longrightarrow> (a, as ! i) \<in> s) }"

declare step_s_mul_ext_def[simp]

text{* The relation step_s_mul_ext is included in s_mul_ext. *}

lemma fold_step:
  assumes "(A,B) \<in> (step_s_mul_ext ns s)"
  shows "(A,B) \<in> (s_mul_ext ns s)"
proof-
from assms obtain a as ps qs where
  A_part: "A = multiset_of ps + {#a#}" and
  B_part: "B = multiset_of qs + multiset_of as" and
  pslqs: "length ps = length qs" and
  ns_ps_qs: "\<forall>i. i < length qs \<longrightarrow> (ps ! i, qs ! i) \<in> ns" and
  s_a_as: "\<forall>i. i < length as \<longrightarrow> (a, as ! i) \<in> s"
unfolding step_s_mul_ext_def by blast
have a_ne: "{#a#} \<noteq> {#}" by simp
{
  fix b assume "b \<in># multiset_of as"
  from in_multiset_of_idx[OF this] obtain i where
    i_bd: "i < length as" and
    eq_b_as: "b = as ! i" by auto
  with s_a_as have "\<exists>c. c \<in># {#a#} \<and> (c,b) \<in> s" by simp
}
hence s_a_as': "\<forall>b. b \<in># multiset_of as \<longrightarrow> (\<exists>c. c \<in># {#a#} \<and> (c,b) \<in> s)" by simp
from A_part and B_part and pslqs and a_ne and ns_ps_qs and s_a_as'
 show ?thesis unfolding s_mul_ext_def by blast
qed

text{* We prove that s_mul_ext can be emulated thanks to step_s_mul_ext.
And conversely. *}

lemma in_single_inversion:
  assumes "x \<in># {#y#}" shows "x = y"
using assms proof (cases "x = y")
 case True thus ?thesis by simp
 next
 case False thus ?thesis using assms by simp
qed

lemma multiset_of_partition:
  "multiset_of (fst (partition P xs)) + multiset_of (snd (partition P xs))
  = multiset_of xs"
proof (induct xs)
case Nil thus ?case by simp
next
case (Cons x xs)
  thus ?case proof (cases "P x")
  case True
    from union_assoc and union_commute
     have "\<And>a b. a + {#x#} + b = (a + b) + {#x#}" by metis
    with Cons and True show ?thesis by simp
  next
  case False with Cons and union_assoc[of _ _ "{#x#}", symmetric]
    show ?thesis by simp
  qed
qed

lemma step_s_mul_ext_emulate_l:
  assumes "refl ns"
  and "(A,B) \<in> (s_mul_ext ns s)"
  shows "(A,B) \<in> (step_s_mul_ext ns s)\<^sup>+"
proof-
from assms(2) obtain as bs A' B' where
  A_part: "A = multiset_of as + A'" and
  B_part: "B = multiset_of bs + B'" and
  aslbs: "length as = length bs" and
  A'_neq: "A' \<noteq> {#}" and
  ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns" and
  s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
thus ?thesis using assms proof (induct A' arbitrary: A B B' as bs)
case empty thus ?case by simp
next
case (add C c) note IH = this
  let ?parts = "partition (\<lambda>x. (c,x) \<in> s) (list_of B')"
  show ?case proof (cases C)
  case empty
    from empty and add(2) have A_part': "A = multiset_of as + {#c#}" by simp
    from multiset_of_list_of[of B', symmetric] and add(3)
     have B_part': "B = multiset_of bs + multiset_of (list_of B')" by simp
    {
      fix i assume i_bd: "i < length (list_of B')"
      from nth_mem_multiset_of[OF i_bd] and multiset_of_list_of[of B']
       have "list_of B' ! i \<in># B'" by simp
      with add(7) obtain a where
        a_in_Cc: "a \<in># C + {#c#}" and 
        s_a_b: "(a,list_of B' ! i) \<in> s" by auto
      from a_in_Cc and empty and in_single_inversion[of c a] and empty_neutral[of "{#c#}"]
       have "a = c" by auto
      with s_a_b have "(c,list_of B' ! i) \<in> s" by simp
    }
    hence s_c_B': "\<forall>i. i < length (list_of B') \<longrightarrow> (c,list_of B' ! i) \<in> s" by simp
    from A_part' and B_part' and add(4) and add(6) and s_c_B'
     have "(A,B) \<in> (step_s_mul_ext ns s)"
     unfolding step_s_mul_ext_def
     by blast
    thus ?thesis by (simp del: step_s_mul_ext_def)
  next
  case (add D d)
    from IH(2) and union_commute and union_lcomm and multiset_of.simps(2)
     have A_part': "A = multiset_of (c # as) + C"
     by metis
    let ?AB = "multiset_of (c # bs) + multiset_of (snd ?parts)"
    from IH(4) have caslcbs: "length (c # as) = length (c # bs)" by simp
    from add have C_ne: "C \<noteq> {#}" by simp
    {
      fix i assume i_bd: "i < length (c # bs)"
      have "((c # as) ! i, (c # bs) ! i) \<in> ns"
      proof (cases i)
      case 0
        from IH(8)[unfolded refl_on_def] and 0
         show ?thesis by simp
      next
      case (Suc j)
        with i_bd have j_bd: "j < length bs" by simp
        from IH(6)[rule_format, OF j_bd] and Suc show ?thesis by simp
      qed
    }
    hence ns_cas_cbs: "\<forall>i. i < length (c # bs) \<longrightarrow> ((c # as) ! i, (c # bs) ! i) \<in> ns" by simp
    {
      fix b assume b_in_2: "b \<in># multiset_of (snd ?parts)"
      have "\<And>x A B C. x \<in># B \<Longrightarrow> A + B = C \<Longrightarrow> x \<in># C" by auto
      with b_in_2 and multiset_of_partition[of "\<lambda>x. (c,x) \<in> s" "list_of B'"]
       and multiset_of_list_of[of B']
       have b_in_B': "b \<in># B'" by simp
       with IH(7) obtain a where
         a_in_Cc: "a \<in># (C + {#c#})"
         and s_a_b: "(a,b) \<in> s" by auto
       have a_in_C: "a \<in># C"
       proof (cases "a = c")
       case False with a_in_Cc show ?thesis by simp
       next
       case True
         from in_multiset_in_set[of "snd ?parts"] and b_in_2 and partition_filter2
          have "b \<in> set (filter (Not \<circ> (\<lambda>x. (c,x) \<in> s)) (list_of B'))" by simp
         with set_filter have "(c,b) \<notin> s" by auto
         with True and s_a_b show ?thesis by simp
       qed
       with s_a_b have "\<exists>a. a \<in># C \<and> (a,b) \<in> s" by fast
     }
     hence s_C_B': "\<forall>b. b \<in># multiset_of (snd ?parts) \<longrightarrow>
       (\<exists>a. a \<in># C \<and> (a,b) \<in> s)" by simp
     from A_part' and caslcbs and C_ne and ns_cas_cbs and s_C_B'
      have s_A_AB: "(A, ?AB) \<in> (s_mul_ext ns s)"
      unfolding s_mul_ext_def by blast
     note LEMMA1 = IH(1)[OF A_part' refl caslcbs C_ne ns_cas_cbs s_C_B'
       assms(1) s_A_AB]
     let ?bs' = "bs @ (snd ?parts)"
     let ?AB' = "multiset_of ?bs' + {#c#}"
     {
       fix A B C :: "'a multiset"
       from union_commute[of B C] and union_assoc[of A B C]
        and union_assoc[of A C B, symmetric]
        have "A + C + B = A + B + C" by auto
     }
     note Hrew = this
     with multiset_of_append[of bs "snd ?parts"] have AB_part: "?AB = ?AB'" by simp
     from Hrew and multiset_of_append[of bs "snd ?parts"] and union_assoc
      have "multiset_of ?bs' + multiset_of (fst ?parts)
         = multiset_of bs + (multiset_of (fst ?parts) + multiset_of (snd ?parts))"
      by simp
     with IH(3) and multiset_of_partition[of "\<lambda>x. (c,x) \<in> s" "list_of B'"]
      and multiset_of_list_of[of B']
      have B_part': "B = multiset_of ?bs' + multiset_of (fst ?parts)" by simp
     from assms(1)[unfolded refl_on_def]
      have ns_bs_bs: "\<forall>i. i < length ?bs' \<longrightarrow> (?bs' ! i, ?bs' ! i) \<in> ns" by simp
     {
       fix i assume i_bd: "i < length (fst ?parts)"
       note b_in_1 = nth_mem_multiset_of[OF i_bd]
       from b_in_1 and partition_filter1
        have "fst ?parts ! i \<in># multiset_of (fst ?parts)" by simp
       with in_multiset_in_set[of _ "fst ?parts ! i"] and set_filter
        have "(c,fst ?parts ! i) \<in> s" by simp
     }
     hence "\<forall>i. i < length (fst ?parts) \<longrightarrow> (c, fst ?parts ! i) \<in> s" by simp
     with AB_part and B_part' and ns_bs_bs
      have LEMMA2: "(?AB,B) \<in> (step_s_mul_ext ns s)"
      unfolding step_s_mul_ext_def by blast
     from trancl_into_trancl[OF LEMMA1 LEMMA2] show ?thesis by assumption
  qed
qed
qed

lemma step_s_mul_ext_emulate_r:
  assumes "trans s"
  and "trans ns"
  and "compatible_l ns s"
  and "compatible_r ns s"
  and "(A,B) \<in> (step_s_mul_ext ns s)\<^sup>+"
  shows "(A,B) \<in> (s_mul_ext ns s)"
proof (induct rule: trancl_induct[OF assms(5)])
case (1 B) with fold_step show ?case by simp
next
case (2 B C)
 with fold_step have s_B_C: "(B,C) \<in> (s_mul_ext ns s)" by simp
 from s_mul_ext_trans[OF tr_ltr[OF assms(1)] tr_ltr[OF assms(2)]
   cl_lcl[OF assms(3)] cr_lcr[OF assms(4)] 2(3) s_B_C] show ?case by assumption
qed

lemma step_s_mul_ext_emulate:
  assumes "refl ns"
  and "trans s"
  and "trans ns"
  and "compatible_l ns s"
  and "compatible_r ns s"
  shows "(step_s_mul_ext ns s)\<^sup>+ = s_mul_ext ns s"
using step_s_mul_ext_emulate_l[OF assms(1)]
and step_s_mul_ext_emulate_r[OF assms(2) assms(3) assms(4) assms(5)]
by fast

lemma SN_s_mul_ext_compat:
  assumes "refl ns"
  and "(A,B) \<in> (s_mul_ext ns s)"
  and "SN_on (step_s_mul_ext ns s) {A}"
  shows "SN_on (step_s_mul_ext ns s) {B}"
  using steps_preserve_SN_on[OF
    trancl_into_rtrancl[OF step_s_mul_ext_emulate_l[OF assms(1,2)]]]
    and assms(3) by blast

lemma permutation_Ex_bij_rev:
  assumes "\<exists>f. bij_betw f {..<length xs} {..<length ys} \<and> (\<forall>i<length xs. xs ! i = ys ! f i)"
  shows "multiset_of xs = multiset_of ys"
using assms proof (induct xs arbitrary: ys)
case Nil
  then obtain f where "bij_betw f {..<length []} {..<length ys}" by auto
  from bij_betw_same_card[OF this] show ?case by simp
next
case (Cons a as)
  from Cons(2) obtain f where
    f_bij: "bij_betw f {..<length (a # as)} {..<length ys}" and
    f_sound: "\<forall>i<length (a # as). (a # as) ! i = ys ! f i" by auto
  from f_bij have f0_bd: "f 0 < length ys"
   unfolding bij_betw_def by auto
  from bij_betw_same_card[OF f_bij] have aaslys: "length (a # as) = length ys" by simp
  have "0 < length (a # as)" by simp
  from f_sound[rule_format, OF this] have eq_a_ys: "a = ys ! f 0" by simp
  let ?g = "\<lambda>n. if f (Suc n) < f 0 then f (Suc n) else f (Suc n) - 1"
  {
    fix x y
    assume x_in_as: "x \<in> {..< length as}"
    and y_in_as: "y \<in> {..< length as}"
    and eq_gx_gy: "?g x = ?g y"
    from y_in_as have Sy_in_aas: "Suc y \<in> {..< length (a # as)}" by simp
    hence Sy_bd: "Suc y < length (a # as)" by simp
    from x_in_as have Sx_in_aas: "Suc x \<in> {..< length (a # as)}" by simp
    hence Sx_bd: "Suc x < length (a # as)" by simp
    have "x = y" proof (cases "f (Suc x) < f 0")
    case True note fSx_bd = this
      show ?thesis proof (cases "f (Suc y) < f 0")
      case True
        with eq_gx_gy and fSx_bd f_bij have eq_fSx_fSy: "f (Suc x) = f (Suc y)" by simp
        from f_bij and eq_fSx_fSy and Sx_in_aas and Sy_in_aas
         show ?thesis unfolding bij_betw_def and inj_on_def by auto
      next
      case False
        with fSx_bd and eq_gx_gy have Heq: "f (Suc x) = f (Suc y) - 1" by simp
        have "Suc y \<noteq> 0" by simp
        with f_bij and Sy_bd have "f (Suc y) \<noteq> f 0"
         unfolding bij_betw_def and inj_on_def by fast
        with Heq and fSx_bd and False show ?thesis by auto
      qed
    next
    case False note fSx_lb = this
      have "Suc x \<noteq> 0" by simp
      with f_bij and Sx_bd have neq_fSx_f0: "f (Suc x) \<noteq> f 0"
       unfolding bij_betw_def and inj_on_def by fast
      with fSx_lb have fSx_pos: "0 < f (Suc x)" by simp
      show ?thesis proof (cases "f (Suc y) < f 0")
      case True
        with fSx_lb and eq_gx_gy have "f (Suc x) - 1 = f (Suc y)" by simp
        with neq_fSx_f0 and True and False show ?thesis by auto
      next
      case False
        with eq_gx_gy and fSx_lb and f_bij
         have Heq: "f (Suc x) - 1 = f (Suc y) - 1" by simp
        have "Suc y \<noteq> 0" by simp
        with f_bij and Sy_bd have neq_fSy_f0: "f (Suc y) \<noteq> f 0"
         unfolding bij_betw_def and inj_on_def by fast
        with False have fSy_pos: "0 < f (Suc y)" by simp
        from Heq and fSx_pos and fSy_pos
         have eq_fSx_fSy: "f (Suc x) = f (Suc y)" by simp
        from f_bij and eq_fSx_fSy and Sx_in_aas and Sy_in_aas
         show ?thesis unfolding bij_betw_def and inj_on_def by auto
      qed
    qed
  }
  hence g_bij1: "inj_on ?g {..< length as}" unfolding inj_on_def by simp
  {
    fix n assume n_in_as: "n \<in> {..< length as}"
    hence Sn_bd: "Suc n < length (a # as)" by simp
    have "?g n \<in> {..< length (remove_nth (f 0) ys)}"
    proof (cases "f (Suc n) < f 0")
    case True
      from f_bij have f0_in: "f 0 \<in> {..<length ys}"
       unfolding bij_betw_def by fast
      hence "f 0 < length ys" by simp
      from f0_in and True and remove_nth_length[OF this]
       show ?thesis by simp
    next
    case False
      have "Suc n \<noteq> 0" by simp
      with f_bij and Sn_bd have neq_fSn_f0: "f (Suc n) \<noteq> f 0"
       unfolding bij_betw_def and inj_on_def by fast
      from f_bij have f0_bd: "f 0 < length ys"
       unfolding bij_betw_def by auto
      from n_in_as have "Suc n < length (a # as)" by simp
      with f_bij have fSn_bd: "f (Suc n) < length ys"
       unfolding bij_betw_def by auto
      with remove_nth_length[OF f0_bd] and False and neq_fSn_f0
       have "f (Suc n) - 1 \<in> {..< length (remove_nth (f 0) ys)}" by force
      with False show ?thesis by simp
    qed
  }
  moreover
  {
    fix n assume n_in: "n \<in> {..< length (remove_nth (f 0) ys)}"
    have "n \<in> ?g ` {..< length as}"
    proof (cases "n < f 0")
    case True
      from n_in and remove_nth_length[OF f0_bd]
       have n_in': "n \<in> {..< length ys}" by simp
      with f_bij have "n \<in> f ` {..< length (a # as)}"
       unfolding bij_betw_def by simp
      then obtain i where
        i_bd: "i < length (a # as)" and
        eq_n_fi: "n = f i" by auto
      show ?thesis proof (cases i)
      case 0 with eq_n_fi and True show ?thesis by simp
      next
      case (Suc j) with eq_n_fi and True and i_bd show ?thesis by simp
      qed
    next
    case False
      from nth_mem_multiset_of[OF f0_bd] have lys_pos: "0 < length ys" by auto
      from n_in and remove_nth_length[OF f0_bd] have Sn_bd: "Suc n < length ys" by auto
      hence "Suc n \<in> {..<length ys}" by simp
      with f_bij have "Suc n \<in> f ` {..<length (a # as)}"
       unfolding bij_betw_def by simp
      then obtain i where
        i_bd: "i < length (a # as)" and
        eq_Sn_fi: "Suc n = f i" by auto
      show ?thesis proof (cases i)
      case 0 with eq_Sn_fi and False show ?thesis by simp
      next
      case (Suc j)
        with eq_Sn_fi and False have "n = ?g j" by simp
        with i_bd and Suc show ?thesis by simp
      qed
    qed
  }
  ultimately have g_bij2: "?g ` {..<length as} = {..<length (remove_nth (f 0) ys)}"
   by blast
  from g_bij1 and g_bij2
  have g_bij: "bij_betw ?g {..<length as} {..<length (remove_nth (f 0) ys)}"
  unfolding bij_betw_def and inj_on_def by simp
  {
    fix i assume i_bd: "i < length as"
    hence Si_bd: "Suc i < length (a # as)" by simp
    have "as ! i = remove_nth (f 0) ys ! ?g i"
    proof (cases "f (Suc i) < f 0")
    case True
      with f_sound[rule_format, OF Si_bd] and remove_nth_sound_l[OF True, of ys]
       show ?thesis by simp
    next
    case False
      have "Suc i \<noteq> 0" by simp
      with f_bij and Si_bd have neq_fSi_f0: "f (Suc i) \<noteq> f 0"
       unfolding bij_betw_def and inj_on_def by fast
      with False have f0_bd: "f 0 \<le> f (Suc i) - 1" by simp
      from False and neq_fSi_f0 have fSi_pos: "0 < f (Suc i)" by simp
      from f_bij and Si_bd have "f (Suc i) < length ys"
       unfolding bij_betw_def by auto
      hence fSi_bd: "f (Suc i) - 1 < length ys" by simp
      with f_sound[rule_format, OF Si_bd] and remove_nth_sound_r[OF f0_bd fSi_bd]
       and False and Suc_diff_1[OF fSi_pos]
       show ?thesis by simp
    qed
  }
  hence g_sound: "\<forall>i<length as. as ! i = remove_nth (f 0) ys ! ?g i" by simp
  from Cons(1) and g_bij and g_sound
   have "multiset_of as = multiset_of (remove_nth (f 0) ys)" by blast
  with remove_nth_soundness[OF f0_bd, symmetric] and eq_a_ys
   and insert_DiffM2[OF nth_mem_multiset_of[OF f0_bd], symmetric]
  show ?case by simp
qed

lemma multiset_of_permut :
  assumes "bij_betw f {..<length as} {..<length as}"
  shows "multiset_of (permut as f) = multiset_of as"
proof -
from permut_length[OF assms] and assms
 have f_bij: "bij_betw f {..<length (permut as f)} {..<length as}" by simp
from permut_sound[of _ as f] and permut_length[OF assms]
 have "\<forall>i. i < length (permut as f) \<longrightarrow> permut as f ! i = as ! f i" by simp
with f_bij and permutation_Ex_bij_rev[of "permut as f" as]
 show ?thesis by fast
qed

lemma SN_ns_mul_ext_compat:
  assumes "refl ns"
  and "compatible_l ns s"
  and "trans ns"
  and "(A,B) \<in> (ns_mul_ext ns s)"
  and "SN_on (step_s_mul_ext ns s) {A}"
  shows "SN_on (step_s_mul_ext ns s) {B}"
proof-
from assms(4) obtain as bs A' B' where
  A_part: "A = multiset_of as + A'" and
  B_part: "B = multiset_of bs + B'" and
  aslbs: "length as = length bs" and
  ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs !i) \<in> ns" and
  s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
show ?thesis proof (cases A')
case (add C c)
  hence A'_ne: "A' \<noteq> {#}" by simp
  from A_part and B_part and aslbs and A'_ne and ns_as_bs and s_A'_B'
   have "(A,B) \<in> (s_mul_ext ns s)" unfolding s_mul_ext_def by blast
  with SN_s_mul_ext_compat and assms show ?thesis by simp
next   
case empty note A'_empty = this
  show ?thesis proof (cases B')
  case (add D d)
    hence "d \<in># B'" by simp
    with s_A'_B' and A'_empty show ?thesis by simp
  next
  case empty
    show ?thesis proof (cases "\<exists>S. S 0 = B \<and> (\<forall>i. (S i, S (Suc i)) \<in> (step_s_mul_ext ns s))")
    case False thus ?thesis by fast
    next
    case True
      then obtain S where
        S0: "S 0 = B" and
        Sn: "\<forall>i. (S i, S (Suc i)) \<in> (step_s_mul_ext ns s)" by fast
      let ?T = "\<lambda>n. if n = 0 then A else S n"
      have T0: "?T 0 = A" by simp
      {
        fix i
        have "(?T i, ?T (Suc i)) \<in> (step_s_mul_ext ns s)"
        proof (cases i)
        case (Suc j) with Sn show ?thesis by simp
        next
        case 0
          from Sn[rule_format, of 0] and S0 obtain x xs ps qs where
            S0_part: "B = multiset_of ps + {#x#}" and
            S1_part: "S 1 = multiset_of qs + multiset_of xs" and
            pslqs: "length ps = length qs" and
            ns_ps_qs: "\<forall>i. i < length qs \<longrightarrow> (ps ! i, qs ! i) \<in> ns" and
            s_a_as: "\<forall>i. i < length xs \<longrightarrow> (x,xs ! i) \<in> s" by auto
          from S0_part and B_part and empty have x_in_bs: "x \<in># multiset_of bs" by simp
          from in_multiset_of_idx[OF this] obtain k where
            k_bd: "k < length bs" and
            eq_x_bs: "x = bs ! k" by auto
          with aslbs have k_bd2: "k < length as" by simp
          from insert_DiffM2[OF nth_mem_multiset_of[OF k_bd2]] and A_part and A'_empty
           and remove_nth_soundness[OF k_bd2]
           have A_part': "A = multiset_of (remove_nth k as) + {#as ! k#}" by simp
          from remove_nth_soundness[OF k_bd] and insert_DiffM2[OF x_in_bs]
           and eq_x_bs have "multiset_of bs = multiset_of (remove_nth k bs) + {#x#}"
           by simp
          with B_part and empty and S0_part
           have eq_bks_ps: "multiset_of (remove_nth k bs) = multiset_of ps" by auto
          from multiset_of_eq_perm and eq_bks_ps[symmetric]
           have "ps <~~> remove_nth k bs" by fast
          from permutation_Ex_bij[OF this] obtain f where
             f_bij: "bij_betw f {..<length ps} {..<length (remove_nth k bs)}" and
             f_sound: "\<forall>i. i < length ps \<longrightarrow> ps ! i = remove_nth k bs ! (f i)"
             by auto
           let ?aks = "permut (remove_nth k as) f"
           from eq_bks_ps and size_multiset_of
            have bkslps: "length (remove_nth k bs) = length ps" by metis
           from remove_nth_length and k_bd and aslbs
            have akslbks: "length (remove_nth k as) = length (remove_nth k bs)" by simp
           with bkslps and pslqs have akslqs: "length (remove_nth k as) = length qs" by simp
           from akslqs and pslqs and akslbks and f_bij
            have f_bij': "bij_betw f {..<length (remove_nth k as)} {..<length (remove_nth k as)}"
            by simp
           from permut_length[OF f_bij'] and akslqs
            have fakslqs: "length ?aks = length qs" by simp
           from multiset_of_permut[OF f_bij'] and A_part'
            have A_part'': "A = multiset_of ?aks + {#as ! k#}" by simp
           {
             fix m assume m_bd: "m < length qs"
             with pslqs have m_bd1: "m < length ps" by simp
             with bkslps have m_bd2: "m < length (remove_nth k bs)" by simp
             with akslbks have m_bd3: "m < length (remove_nth k as)" by simp
             with f_bij' have fm_bd: "f m < length (remove_nth k as)"
              unfolding bij_betw_def by fast
             with akslbks have fm_bd': "f m < length (remove_nth k bs)" by simp
             from remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, rule_format, OF fm_bd']
              and permut_sound[OF m_bd3]
              have ns_faks_bks: "(?aks ! m, remove_nth k bs ! f m) \<in> ns" by simp
             from ns_ps_qs[rule_format, OF m_bd] and f_sound[rule_format, OF m_bd1]
              and `trans ns`[unfolded trans_def, rule_format, OF this]
              have "(?aks ! m, qs ! m) \<in> ns" by simp
           }
           hence ns_faks_qs: "\<forall>i. i < length qs \<longrightarrow> (?aks ! i, qs ! i) \<in> ns" by simp
           {
             fix m assume m_bd: "m < length xs"
             from eq_x_bs and s_a_as[rule_format, OF m_bd] and `compatible_l ns s`
              [unfolded compatible_l_def] ns_as_bs[rule_format, OF k_bd]
             have "(as ! k, xs ! m) \<in> s" by auto
           }
           hence s_ask_xs: "\<forall>i. i < length xs \<longrightarrow> (as ! k, xs ! i) \<in> s" by simp
           from A_part'' and S1_part and fakslqs and ns_faks_qs and s_ask_xs and 0
           show ?thesis unfolding step_s_mul_ext_def by auto
         qed
       }
       with T0 and assms(5) have "False" by fast
       thus ?thesis by simp
    qed
  qed
qed
qed

lemma SN_step_s_mul_ext_strong:
  assumes "refl ns"
  and "compatible_l ns s"
  and "trans ns"
  and "\<forall>y. y \<in># M \<longrightarrow> SN_on s {y}"        (is "?H M")
  shows "SN_on (step_s_mul_ext ns s) {M}" (is "SN_on ?R {M}")
proof -
  let ?Sn = "SN_on ?R"
  {
    fix M a
    assume Sn_M: "?Sn {M}"
    and IH1: "\<forall>M0 b. (M,M0) \<in> ?R \<longrightarrow> (a,b) \<in> ns \<longrightarrow> ?Sn {M0 + {#b#}}"
    and IH2: "\<forall>b. (a,b) \<in> s \<longrightarrow> (\<forall>M0. ?Sn {M0} \<longrightarrow> ?Sn {M0 + {#b#}})"
    have "\<forall>b. (a,b) \<in> ns \<longrightarrow> ?Sn {M + {#b#}}"
    proof -
    {
      fix b assume ns_a_b: "(a,b) \<in> ns"
      fix M0 assume s_M_M0: "(M + {#b#},M0) \<in> ?R"
      then obtain x xs ps qs where
        M_part: "M + {#b#} = multiset_of ps + {#x#}" and
        M0_part: "M0 = multiset_of qs + multiset_of xs" and
        pslqs: "length ps = length qs" and
        ns_ps_qs: "\<forall>i. i < length qs \<longrightarrow> (ps ! i, qs ! i) \<in> ns" and
        s_x_xs: "\<forall>i. i < length xs \<longrightarrow> (x,xs ! i) \<in> s" by auto
      {
        assume "b \<in># {#x#}"
        with in_single_inversion have eq_b_x: "b = x" by simp
        have "?Sn {M0}"
        using M0_part and s_x_xs proof (induct xs arbitrary: M0)
        case Nil note M0_part' = this
          from M_part and eq_b_x have M_part': "M = multiset_of ps + multiset_of []" by simp
          have s_e_e: "\<forall>b. b \<in># multiset_of [] \<longrightarrow> (\<exists>a. a \<in># multiset_of [] \<and> (a,b) \<in> s)" by simp
          from M_part' and M0_part' and pslqs and ns_ps_qs and s_e_e
           have "(M, M0) \<in> (ns_mul_ext ns s)" unfolding ns_mul_ext_def by blast
          with SN_ns_mul_ext_compat and Sn_M and assms show ?case by simp
        next
        case (Cons c cs)
          hence c_in_M0: "c \<in># M0" by simp
          let ?M0' = "multiset_of qs + multiset_of cs"
          from Cons(2) and union_assoc and multiset_of.simps(2)
           have M0_part': "M0 = ?M0' + {#c#}" by metis
          hence M0c: "M0 - {#c#} = ?M0'" by simp
          {
            fix i assume "i < length cs"
            hence "Suc i < length (c # cs)" by simp
            from Cons(3)[rule_format, OF this] have "(x, cs ! i) \<in> s" by simp
          }
          hence s_x_cs: "\<forall>i. i < length cs \<longrightarrow> (x,cs ! i) \<in> s" by simp
          have "0 < length (c # cs)" by simp
          from Cons(3)[rule_format, OF this] and eq_b_x have "(b, c) \<in> s" by simp
          with ns_a_b assms(2)[unfolded compatible_l_def] have "(a,c) \<in> s" by auto
          from IH2[rule_format, OF this Cons(1)[OF M0c s_x_cs]]
             and insert_DiffM2[OF c_in_M0]
           show ?case by auto
        qed
      }
      note SC1 = this
      {
        assume a_in_ps: "b \<in># multiset_of ps"
        from in_multiset_of_idx[OF a_in_ps] obtain m where
          m_bd: "m < length ps" and
          eq_b_ps: "b = ps ! m" by fast
        with pslqs have m_bd': "m < length qs" by simp
        from M0_part and nth_mem_multiset_of[OF m_bd'] have qs_in_M0: "qs ! m \<in># M0" by simp
        from eq_b_ps and diff_union_swap_multiset[OF nth_mem_multiset_of[OF m_bd], of "{#x#}"]
         and remove_nth_soundness[OF m_bd] and M_part
         have "M + {#b#} - {#b#} = multiset_of (remove_nth m ps) + {#x#}" by simp
        hence M_part': "M = multiset_of (remove_nth m ps) + {#x#}" by simp
        from diff_union_swap_multiset[OF nth_mem_multiset_of[OF m_bd'], of "multiset_of xs"]
         and remove_nth_soundness[OF m_bd'] and M0_part
         have M0_part': "M0 - {#qs ! m#} = multiset_of (remove_nth m qs) + multiset_of xs" by simp
        from pslqs and remove_nth_length[OF m_bd] and remove_nth_length[OF m_bd']
         have pmslqms: "length (remove_nth m ps) = length (remove_nth m qs)" by simp
        from M_part' and M0_part' and remove_nth_P_compat[OF pslqs[symmetric] ns_ps_qs, of m]
         and s_x_xs and pmslqms have T: "(M, M0-{#qs ! m#}) \<in> ?R"
         unfolding step_s_mul_ext_def by blast
        from assms(3)[unfolded trans_def, rule_format, OF ns_a_b] and eq_b_ps and
         ns_ps_qs[rule_format, OF m_bd'] have "(a,qs ! m) \<in> ns" by simp
        from IH1[rule_format, OF T this] and insert_DiffM2[OF qs_in_M0]
         have "?Sn {M0}" by auto
      }
      note SC2 = this
      have "b \<in># M + {#b#}" by simp
      with SC1 and SC2 and M_part have "?Sn {M0}" by auto
    }
    hence C: "\<forall>b. (a,b) \<in> ns \<longrightarrow> (\<forall>M0. (M + {#b#}, M0) \<in> ?R \<longrightarrow> SN_on ?R {M0})" by simp
    {
      fix b assume ns_a_b: "(a,b) \<in> ns"
      assume Hf: "\<exists>S. S 0 = M + {#b#} \<and> (\<forall>i. (S i, S (Suc i)) \<in> step_s_mul_ext ns s)"
      then obtain S where
        S0: "S 0 = M + {#b#}" and
        SN: "\<forall>i. (S i, S (Suc i)) \<in> step_s_mul_ext ns s" by fast
      let ?T = "\<lambda>n. S (Suc n)"
      from S0 and SN[rule_format, of 0] have H1: "(M + {#b#}, ?T 0) \<in> step_s_mul_ext ns s" 
       by simp
      from SN have TN: "\<forall>i. (?T i, ?T (Suc i)) \<in> step_s_mul_ext ns s" by auto
      hence "\<exists>T. T 0 = ?T 0 \<and> (\<forall>i. (T i, T (Suc i)) \<in> step_s_mul_ext ns s)" by auto
      with C[unfolded SN_on_def, rule_format, OF ns_a_b H1] and bot_nat_def
       have "False" by blast
    }
    thus ?thesis unfolding SN_on_def by force
    qed
  }
  note astuce = this
  {
    fix M assume "?H M"
    have "?Sn {M}"
    using `?H M` proof (induct M)
    case empty
      {
        assume "\<exists>S. S 0 = {#} \<and> (\<forall>i. (S i, S (Suc i)) \<in> ?R)"
        then obtain S where
          S0: "S 0 = {#}" and
          SN: "\<forall>i. (S i, S (Suc i)) \<in> (step_s_mul_ext ns s)" by auto
        from S0 and SN[rule_format, of 0] have "({#}, S 1) \<in> step_s_mul_ext ns s" by simp
        then obtain x ps where
          "{#} = multiset_of ps + {#x#}" by auto
        hence False by auto
      }
      thus ?case unfolding SN_on_def not_def by blast
    next
    case (add M a)
      hence Hadd: "\<forall>y. y \<in># M \<longrightarrow> SN_on s {y}" and SNa: "SN_on s {a}" by auto
      from SNa have "\<forall>M. ?Sn {M} \<longrightarrow> ?Sn {M + {#a#}}" (is "?P a")
      proof (induct rule: SN_on_induct_acc_style)
      case (IH a) note hyps = this
        from hyps(1) and step_preserves_SN_on
         have "\<forall>b. (a,b) \<in> s \<longrightarrow> SN_on s {b}" by auto
        with hyps(2) have H: "\<forall>b. (a,b) \<in> s \<longrightarrow> (\<forall>M0. ?Sn {M0} \<longrightarrow> ?Sn {M0 + {#b#}})" by simp
        {
          fix M assume "?Sn {M}"
          hence "\<forall>b. (a, b) \<in> ns \<longrightarrow> ?Sn {M + {#b#}}"
            by (induct M rule: SN_on_induct_acc_style, intro astuce[OF _ _ H], auto)
        } note T = this
        from assms(1)[unfolded refl_on_def] have "(a,a) \<in> ns" by simp
        from T[rule_format, OF _ this]
         show ?case by simp
      qed
      from this[rule_format, OF add(1)] and Hadd show ?case by simp
    qed
  } note main = this
  show ?thesis 
    by (rule main, rule assms(4))
qed

lemma SN_step_s_mul_ext:
  assumes "refl ns"
  and "compatible_l ns s"
  and "trans ns"
  and "SN s"
  shows "SN (step_s_mul_ext ns s)"
proof -
  from assms(4) have "\<And>M. \<forall>y. y \<in># M \<longrightarrow> SN_on s {y}" unfolding SN_on_def by auto
  from SN_step_s_mul_ext_strong[OF assms(1-3) this] show ?thesis unfolding SN_on_def by blast
qed


lemma SN_s_mul_ext_strong:
  assumes "order_pair s ns"
  and "\<forall>y. y \<in># M \<longrightarrow> SN_on s {y}"
  shows "SN_on (s_mul_ext ns s) {M}"
proof -
  interpret order_pair s ns by fact
  have ass: "refl ns" "trans ns" "compatible_l ns s" "compatible_r ns s" "trans s"
    unfolding compatible_l_def compatible_r_def using trans_S trans_NS compat_NS_S compat_S_NS refl_NS by auto
  show ?thesis 
    using SN_on_trancl
      and step_s_mul_ext_emulate[OF ass(1) ass(5) ass(2) ass(3) ass(4)]
    and SN_step_s_mul_ext_strong[OF ass(1) ass(3) ass(2) assms(2)]
    by metis
qed

lemma (in order_pair) mul_ext_order_pair: "order_pair (s_mul_ext NS S) (ns_mul_ext NS S)" (is "order_pair ?S ?NS")
proof -
  note c = tr_ltr[OF trans_S] tr_ltr[OF trans_NS] 
    cl_lcl[unfolded compatible_l_def, OF compat_NS_S]
    cr_lcr[unfolded compatible_r_def, OF compat_S_NS]
  show ?thesis
  proof
    from s_mul_ext_trans[OF c]
    show "trans ?S" unfolding trans_def by blast
  next
    from ns_mul_ext_trans[OF c]
    show "trans ?NS" unfolding trans_def by blast
  next
    from ns_s_mul_ext_trans[OF c]
    show "?NS O ?S \<subseteq> ?S" by blast
  next
    from s_ns_mul_ext_trans[OF c]
    show "?S O ?NS \<subseteq> ?S" by blast
  next
    from ns_mul_ext_refl[OF r_lr[OF refl_NS]]
    show "refl ?NS" unfolding refl_on_def by auto
  qed
qed

lemma (in SN_order_pair) mul_ext_SN_order_pair: "SN_order_pair (s_mul_ext NS S) (ns_mul_ext NS S)"
  (is "SN_order_pair ?S ?NS")
proof -
  have ass: "refl NS" "trans NS" "compatible_l NS S" "compatible_r NS S" "trans S" "SN S"
    unfolding compatible_l_def compatible_r_def using trans_S trans_NS compat_NS_S compat_S_NS refl_NS SN by auto
  from mul_ext_order_pair
  interpret order_pair ?S ?NS .
  show ?thesis
  proof
    from SN_step_s_mul_ext ass
      and SN_imp_SN_trancl and SN_step_s_mul_ext
      and step_s_mul_ext_emulate 
    show "SN ?S" by metis
  qed
qed

lemma mul_ext_compat:
  assumes compat: "\<And> s t u. \<lbrakk>s \<in> set ss; t \<in> set ts; u \<in> set us\<rbrakk> \<Longrightarrow>
    (snd (f s t) \<and> fst (f t u) \<longrightarrow> fst (f s u)) \<and> 
    (fst (f s t) \<and> snd (f t u) \<longrightarrow> fst (f s u)) \<and> 
    (snd (f s t) \<and> snd (f t u) \<longrightarrow> snd (f s u)) \<and>
    (fst (f s t) \<and> fst (f t u) \<longrightarrow> fst (f s u))"
  shows "
    (snd (mul_ext f ss ts) \<and> fst (mul_ext f ts us) \<longrightarrow> fst (mul_ext f ss us)) \<and> 
    (fst (mul_ext f ss ts) \<and> snd (mul_ext f ts us) \<longrightarrow> fst (mul_ext f ss us)) \<and> 
    (snd (mul_ext f ss ts) \<and> snd (mul_ext f ts us) \<longrightarrow> snd (mul_ext f ss us)) \<and>
    (fst (mul_ext f ss ts) \<and> fst (mul_ext f ts us) \<longrightarrow> fst (mul_ext f ss us)) "
proof (intro conjI)
  let ?s = "{(t,u). fst (f t u)}"
  let ?ns = "{(t,u). snd (f t u)}"
  let ?mss = "multiset_of ss"
  let ?mts = "multiset_of ts"
  let ?mus = "multiset_of us"
  from assms have lt_s: "locally_trans ?s ?mss ?mts ?mus"
    unfolding locally_trans_def in_multiset_in_set by auto
  from assms have lt_ns: "locally_trans ?ns ?mss ?mts ?mus"
    unfolding locally_trans_def in_multiset_in_set by auto  
  from assms have lcl: "locally_compatible_l ?ns ?s ?mss ?mts ?mus"
    unfolding locally_compatible_l_def in_multiset_in_set by auto
  from assms have lcr: "locally_compatible_r ?ns ?s ?mss ?mts ?mus"
    unfolding locally_compatible_r_def in_multiset_in_set by auto
  from s_mul_ext_trans[OF lt_s lt_ns lcl lcr]
    show "fst (mul_ext f ss ts) \<and> fst (mul_ext f ts us) \<longrightarrow> fst (mul_ext f ss us)"
    unfolding mul_ext_def by simp
  from ns_mul_ext_trans[OF lt_s lt_ns lcl lcr]
    show "snd (mul_ext f ss ts) \<and> snd (mul_ext f ts us) \<longrightarrow> snd (mul_ext f ss us)"
    unfolding mul_ext_def by simp
  from ns_s_mul_ext_trans[OF lt_s lt_ns lcl lcr]
    show "snd (mul_ext f ss ts) \<and> fst (mul_ext f ts us) \<longrightarrow> fst (mul_ext f ss us)"
    unfolding mul_ext_def by simp
  from s_ns_mul_ext_trans[OF lt_s lt_ns lcl lcr]
    show "fst (mul_ext f ss ts) \<and> snd (mul_ext f ts us) \<longrightarrow> fst (mul_ext f ss us)"
    unfolding mul_ext_def by simp
qed

lemma mul_ext_cong[fundef_cong]:
  assumes "multiset_of xs1 = multiset_of ys1"
  and "multiset_of xs2 = multiset_of ys2"
  and "\<And> x x'. x \<in> set ys1 \<Longrightarrow> x' \<in> set ys2 \<Longrightarrow> f x x' = g x x'"
  shows "mul_ext f xs1 xs2 = mul_ext g ys1 ys2"
proof -
let ?s1 = "{(x,y). fst (f x y)}" and ?ns1 = "{(x,y). snd (f x y)}"
and ?s2 = "{(x,y). fst (g x y)}" and ?ns2 = "{(x,y). snd (g x y)}"
{
  fix y y' assume y_in_ys1: "y \<in> set ys1" and y_in_ys2: "y' \<in> set ys2"
  from assms(3)[OF y_in_ys1 y_in_ys2]
  have "(y,y') \<in> ?s1 \<longleftrightarrow> (y,y') \<in> ?s2" by simp
}
hence H1: "\<forall>y y'. y \<in> set ys1 \<longrightarrow> y' \<in> set ys2 \<longrightarrow> (y,y') \<in> ?s1 \<longleftrightarrow> (y,y') \<in> ?s2" by simp
{
  fix y y' assume y_in_ys1: "y \<in> set ys1" and y_in_ys2: "y' \<in> set ys2"
  from assms(3)[OF y_in_ys1 y_in_ys2]
  have "(y,y') \<in> ?ns1 \<longleftrightarrow> (y,y') \<in> ?ns2" by simp
}
hence H2: "\<forall>y y'. y \<in> set ys1 \<longrightarrow> y' \<in> set ys2 \<longrightarrow> (y,y') \<in> ?ns1 \<longleftrightarrow> (y,y') \<in> ?ns2" by simp
from s_mul_ext_extensional[OF assms(1) assms(2) H2 H1]
 and ns_mul_ext_extensional[OF assms(1) assms(2) H2 H1]
 show ?thesis unfolding mul_ext_def by simp
qed

lemma all_nstri_imp_mul_nstri:
  assumes "\<forall>i<length ys. snd (f (xs ! i) (ys ! i))"
  and "length xs = length ys"
  shows "snd (mul_ext f xs ys)"
proof-
from assms(1) have "\<forall>i. i < length ys \<longrightarrow> (xs ! i, ys ! i) \<in> {(x,y). snd (f x y)}" by simp
from all_ns_ns_mul_ext[OF assms(2) this] show ?thesis by (simp add: mul_ext_def)
qed

lemma relation_inter:
  shows "{(x,y). P x y} \<inter> {(x,y). Q x y} = {(x,y). P x y \<and> Q x y}"
by blast

lemma mul_ext_unfold:
 "(x,y) \<in> {(a,b). fst (mul_ext g a b)} \<longleftrightarrow> (multiset_of x, multiset_of y) \<in> (s_mul_ext {(a,b). snd (g a b)} {(a,b). fst (g a b)})"
using assms unfolding mul_ext_def by simp

lemma mul_ext_SN:
  assumes "\<forall>x. snd (g x x)"
  and "\<forall> x y z. fst (g x y) \<longrightarrow> snd (g y z) \<longrightarrow> fst (g x z)"
  and "\<forall> x y z. snd (g x y) \<longrightarrow> fst (g y z) \<longrightarrow> fst (g x z)"
  and "\<forall> x y z. snd (g x y) \<longrightarrow> snd (g y z) \<longrightarrow> snd (g x z)"
  and "\<forall>x y z. fst (g x y) \<longrightarrow> fst (g y z) \<longrightarrow> fst (g x z)"
  shows "SN {(ys, xs).
  (\<forall>y\<in>set ys. SN_on {(s ,t). fst (g s t)} {y}) \<and>
  fst (mul_ext g ys xs)}"
proof-
  let ?R1 = "\<lambda> xs ys. \<forall>y\<in> set ys. SN_on {(s ,t). fst (g s t)} {y}"
  let ?R2 = "\<lambda> xs ys. fst (mul_ext g ys xs)"
  let ?s = "{(x,y). fst (g x y)}"
  let ?ns = "{(x,y). snd (g x y)}"
  have op: "order_pair ?s ?ns"
  proof
    from assms(1) show "refl ?ns" unfolding refl_on_def by auto next
    from assms(2) show "?s O ?ns \<subseteq> ?s" by auto next
    from assms(3) show "?ns O ?s \<subseteq> ?s" by auto next
    from assms(4) show "trans ?ns" unfolding trans_def by blast next  
    from assms(5) show "trans ?s" unfolding trans_def by blast
  qed
  let ?R = "{(ys, xs). ?R1 xs ys \<and> ?R2 xs ys}"
  let ?Sn = "SN_on ?R"
  {
    fix ys xs
    assume R_ys_xs: "(ys, xs) \<in> ?R"
    let ?mys = "multiset_of ys"
    let ?mxs = "multiset_of xs"
    from R_ys_xs have  HSN_ys: "\<forall>y. y \<in> set ys \<longrightarrow> SN_on ?s {y}" by simp
    with in_multiset_in_set[of ys] have "\<forall>y. y \<in># ?mys \<longrightarrow> SN_on ?s {y}" by simp
    from SN_s_mul_ext_strong[OF op this] and mul_ext_unfold
    have "SN_on {(ys,xs). fst (mul_ext g ys xs)} {ys}" by fast
    from relation_inter[of ?R2 ?R1] and SN_on_weakening[OF this]
    have "SN_on ?R {ys}" by blast
  }
  hence Hyp: "\<forall>ys xs. (ys,xs) \<in> ?R \<longrightarrow> SN_on ?R {ys}" by auto
  {
    fix ys
    have "SN_on ?R {ys}"
    proof (cases "\<exists> xs. (ys, xs) \<in> ?R")
      case True with Hyp show ?thesis by simp
    next
      case False thus ?thesis by auto
    qed
  }
  thus ?thesis unfolding SN_on_def by simp
qed

lemma mul_ext_stri_imp_nstri:
  assumes "fst (mul_ext f as bs)"
  shows "snd (mul_ext f as bs)"
  using assms and s_ns_mul_ext unfolding mul_ext_def by auto

lemma ns_mul_ext_diff_single_r:
  assumes "(A,B) \<in> (ns_mul_ext ns s)"
  shows "(A,B - {#x#}) \<in> (ns_mul_ext ns s)"
proof (cases "x \<in># B")
  case False
  from assms and diff_single_trivial[OF this] show ?thesis by simp
next
  case True
  from assms obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
  {
    assume "x \<in># B'"
    with B_part and diff_union_single_conv and union_commute
    have B_part2: "B - {#x#} = multiset_of bs + (B' - {#x#})" by metis
    from s_A'_B' have "\<forall>b. b \<in># (B' - {#x#}) \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
    with A_part and B_part2 and aslbs and ns_as_bs
    have "(A,(B - {#x#})) \<in> (ns_mul_ext ns s)" unfolding ns_mul_ext_def by blast
  }
  moreover
  {
    assume "x \<in># multiset_of bs"
    from in_multiset_of_idx[OF this] obtain i where
      i_bd: "i < length bs" and
      eq_x_bs: "x = bs ! i" by auto
    with aslbs have i_bd2: "i < length as" by simp
    from s_A'_B' have s_A'i_B':  "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># (A' + {#as ! i#}) \<and> (a,b) \<in> s)"
      by auto
    note ns_ais_bis = remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of i]
    from remove_nth_soundness[OF i_bd] and nth_mem_multiset_of[OF i_bd]
      and diff_union_swap_multiset and B_part and eq_x_bs
    have B_part2: "B - {#x#} = multiset_of (remove_nth i bs) + B'" by metis
    from remove_nth_soundness[OF i_bd2] and union_assoc and nth_mem_multiset_of[OF i_bd2]
      union_commute and A_part and insert_DiffM2[of  "multiset_of as" "as ! i"]
    have A_part2: "A = multiset_of (remove_nth i as) + (A' + {#as ! i#})" by metis
    from remove_nth_length[OF i_bd] and remove_nth_length[OF i_bd2] and aslbs
    have aislbis: "length (remove_nth i as) = length (remove_nth i bs)" by simp
    from A_part2 and B_part2 and aislbis and ns_ais_bis and s_A'i_B'
    have "(A,(B - {#x#})) \<in> (ns_mul_ext ns s)" unfolding ns_mul_ext_def by blast
  }
  ultimately show ?thesis using B_part and True by force
qed

lemma ns_mul_ext_diff_multiset_r:
  assumes "(A,B) \<in> (ns_mul_ext ns s)"
  shows "(A,B - C) \<in> (ns_mul_ext ns s)"
  using assms
proof (induct C arbitrary: A B)
  case empty
  thus ?case by auto
next
  case (add C c) note IH = this
  with ns_mul_ext_diff_single_r have "(A,(B - {#c#})) \<in> (ns_mul_ext ns s)" by simp
  with IH(1) have "(A, (B - {#c#} - C)) \<in> (ns_mul_ext ns s)" by simp
  with diff_add and union_commute show ?case by metis
qed

lemma s_mul_ext_diff_single_r:
  assumes "(A,B) \<in> (s_mul_ext ns s)"
  shows "(A,B - {#x#}) \<in> (s_mul_ext ns s)"
proof (cases "x \<in># B")
  case False
  from assms and diff_single_trivial[OF this] show ?thesis by simp
next
  case True
  from assms obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    A'_ne: "A' \<noteq> {#}" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
  {
    assume "x \<in># B'"
    with B_part and diff_union_single_conv and union_commute
    have B_part2: "B - {#x#} = multiset_of bs + (B' - {#x#})" by metis
    from s_A'_B' have "\<forall>b. b \<in># (B' - {#x#}) \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
    with A_part and B_part2 and aslbs and A'_ne and ns_as_bs
    have "(A,(B - {#x#})) \<in> (s_mul_ext ns s)" unfolding s_mul_ext_def by blast
  }
  moreover
  {
    assume "x \<in># multiset_of bs"
    from in_multiset_of_idx[OF this] obtain i where
      i_bd: "i < length bs" and
      eq_x_bs: "x = bs ! i" by auto
    with aslbs have i_bd2: "i < length as" by simp
    from s_A'_B' have s_A'i_B':  "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># (A' + {#as ! i#}) \<and> (a,b) \<in> s)"
      by auto
    note ns_ais_bis = remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of i]
    have A'i_ne: "A' + {#as ! i#} \<noteq> {#}" by simp
    from remove_nth_soundness[OF i_bd] and nth_mem_multiset_of[OF i_bd]
      and diff_union_swap_multiset and B_part and eq_x_bs
    have B_part2: "B - {#x#} = multiset_of (remove_nth i bs) + B'" by metis
    from remove_nth_soundness[OF i_bd2] and union_assoc and nth_mem_multiset_of[OF i_bd2]
      union_commute and A_part and insert_DiffM2[of  "multiset_of as" "as ! i"]
    have A_part2: "A = multiset_of (remove_nth i as) + (A' + {#as ! i#})" by metis
    from remove_nth_length[OF i_bd] and remove_nth_length[OF i_bd2] and aslbs
    have aislbis: "length (remove_nth i as) = length (remove_nth i bs)" by simp
    from A_part2 and B_part2 and aislbis and A'i_ne and ns_ais_bis and s_A'i_B'
    have "(A,(B - {#x#})) \<in> (s_mul_ext ns s)" unfolding s_mul_ext_def by blast
  }
  ultimately show ?thesis using B_part and True by force
qed

lemma s_mul_ext_diff_multiset_r:
  assumes "(A,B) \<in> (s_mul_ext ns s)"
  shows "(A,B - C) \<in> (s_mul_ext ns s)"
  using assms
proof (induct C arbitrary: A B)
  case empty
  thus ?case by auto
next
  case (add C c) note IH = this
  with s_mul_ext_diff_single_r have "(A,(B - {#c#})) \<in> (s_mul_ext ns s)" by simp
  with IH(1) have "(A, (B - {#c#} - C)) \<in> (s_mul_ext ns s)" by simp
  with diff_add and union_commute show ?case by metis
qed

lemma ns_ns_mul_ext_union_compat:
  assumes "(A,B) \<in> ns_mul_ext ns s"
  and "(C,D) \<in> ns_mul_ext ns s"
  shows "(A + C, B + D) \<in> ns_mul_ext ns s"
proof-
  from assms(1) obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)"
    by auto
  from assms(2) obtain cs ds C' D' where
    C_part: "C = multiset_of cs + C'" and
    D_part: "D = multiset_of ds + D'" and
    cslds: "length cs = length ds" and
    ns_cs_ds: "\<forall>i. i < length ds \<longrightarrow> (cs ! i, ds ! i) \<in> ns" and
    s_C'_D': "\<forall>d. d \<in># D' \<longrightarrow> (\<exists>c. c \<in># C' \<and> (c,d) \<in> s)"
    by auto
  from A_part and C_part and multiset_of_append and union_commute and union_assoc
  have AC_part: "A + C = multiset_of (as @ cs) + (A' + C')" by metis
  from B_part and D_part and multiset_of_append and union_commute and union_assoc
  have BD_part: "B + D = multiset_of (bs @ ds) + (B' + D')" by metis
  from aslbs and cslds and length_append
  have acslbds: "length (as @ cs) = length (bs @ ds)" by simp
  {
    fix i assume i_bd: "i < length (bs @ ds)"
    have "((as @ cs) ! i, (bs @ ds) ! i) \<in> ns"
    proof (cases "i < length bs")
      case True
      with aslbs and nth_append and ns_as_bs[rule_format, OF True]
      show ?thesis by metis
    next
      case False
      with aslbs have H: "\<not> i < length as" by simp
      from False and i_bd and length_append
      have i_bd2: "i - length bs < length ds" by simp
      from ns_cs_ds[rule_format, OF this] and aslbs
      have "(cs ! (i - length as), ds ! (i - length bs)) \<in> ns" by simp
      with H and False and nth_append show ?thesis by metis
    qed
  }
  hence ns_acs_bds: "\<forall>i. i < length (bs @ ds) \<longrightarrow> ((as @ cs) ! i, (bs @ ds) ! i) \<in> ns" by simp
  {
    fix b assume b_in: "b \<in># (B' + D')"
    have "\<exists>a. a \<in># (A' + C') \<and> (a,b) \<in> s"
    proof (cases "b \<in># B'")
      case True
      from s_A'_B'[rule_format, OF this] obtain a where "a \<in># A'" and "(a,b) \<in> s" by auto
      thus ?thesis by auto
    next
      case False
      with b_in have "b \<in># D'" by simp
      from s_C'_D'[rule_format, OF this] obtain a where "a \<in># C'" and "(a,b) \<in> s" by auto
      thus ?thesis by auto
    qed
  }
  with AC_part and BD_part and acslbds and ns_acs_bds
  show ?thesis unfolding ns_mul_ext_def by blast
qed


lemma s_ns_mul_ext_union_compat:
  assumes "(A,B) \<in> s_mul_ext ns s"
  and "(C,D) \<in> ns_mul_ext ns s"
  shows "(A + C, B + D) \<in> s_mul_ext ns s"
proof-
  from assms(1) obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    A'_ne: "A' \<noteq> {#}" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)"
    by auto
  from assms(2) obtain cs ds C' D' where
    C_part: "C = multiset_of cs + C'" and
    D_part: "D = multiset_of ds + D'" and
    cslds: "length cs = length ds" and
    ns_cs_ds: "\<forall>i. i < length ds \<longrightarrow> (cs ! i, ds ! i) \<in> ns" and
    s_C'_D': "\<forall>d. d \<in># D' \<longrightarrow> (\<exists>c. c \<in># C' \<and> (c,d) \<in> s)"
    by auto
  from A_part and C_part and multiset_of_append and union_commute and union_assoc
  have AC_part: "A + C = multiset_of (as @ cs) + (A' + C')" by metis
  from B_part and D_part and multiset_of_append and union_commute and union_assoc
  have BD_part: "B + D = multiset_of (bs @ ds) + (B' + D')" by metis
  from aslbs and cslds and length_append
  have acslbds: "length (as @ cs) = length (bs @ ds)" by simp
  from A'_ne have A'C'_ne: "A' + C' \<noteq> {#}" by simp
  {
    fix i assume i_bd: "i < length (bs @ ds)"
    have "((as @ cs) ! i, (bs @ ds) ! i) \<in> ns"
    proof (cases "i < length bs")
      case True
      with aslbs and nth_append and ns_as_bs[rule_format, OF True]
      show ?thesis by metis
    next
      case False
      with aslbs have H: "\<not> i < length as" by simp
      from False and i_bd and length_append
      have i_bd2: "i - length bs < length ds" by simp
      from ns_cs_ds[rule_format, OF this] and aslbs
      have "(cs ! (i - length as), ds ! (i - length bs)) \<in> ns" by simp
      with H and False and nth_append show ?thesis by metis
    qed
  }
  hence ns_acs_bds: "\<forall>i. i < length (bs @ ds) \<longrightarrow> ((as @ cs) ! i, (bs @ ds) ! i) \<in> ns" by simp
  {
    fix b assume b_in: "b \<in># (B' + D')"
    have "\<exists>a. a \<in># (A' + C') \<and> (a,b) \<in> s"
    proof (cases "b \<in># B'")
      case True
      from s_A'_B'[rule_format, OF this] obtain a where "a \<in># A'" and "(a,b) \<in> s" by auto
      thus ?thesis by auto
    next
      case False
      with b_in have "b \<in># D'" by simp
      from s_C'_D'[rule_format, OF this] obtain a where "a \<in># C'" and "(a,b) \<in> s" by auto
      thus ?thesis by auto
    qed
  }
  with AC_part and BD_part and acslbds and A'C'_ne and ns_acs_bds
  show ?thesis unfolding s_mul_ext_def by blast
qed

lemma mul_ext_singletons:
  assumes "\<forall> a b. fst (f a b) \<longrightarrow> snd (f a b)"
  shows "mul_ext f [a] [b] = f a b"
proof (cases "f a b")
  let ?s = "{(x,y). fst (f x y)}"
  let ?ns = "{(x,y). snd (f x y)}"
  case (Pair s ns)
  show ?thesis
  proof (cases s)
    case True
    with Pair have fst_fab: "fst (f a b)" by simp
    with all_s_s_mul_ext[of "{#a#}" "{#b#}" ?s ?ns]
    have "({#a#}, {#b#}) \<in> s_mul_ext ?ns ?s" by auto
    with s_ns_mul_ext[OF this] and assms[rule_format, OF fst_fab] and fst_fab
      and multiset_of.simps and pair_collapse[of "f a b"]
    show ?thesis unfolding mul_ext_def by simp
  next
    case False
    with Pair have nfst_fab: "\<not> fst (f a b)" by simp
    {
      assume "fst (mul_ext f [a] [b])"
      then obtain ps qs P' Q' where
        P_part: "{#a#} = multiset_of ps + P'" and
        Q_part: "{#b#} = multiset_of qs + Q'" and
        pslqs: "length ps = length qs" and
        P'_ne: "P' \<noteq> {#}" and
        s_P'_Q': "\<forall>q. q \<in># Q' \<longrightarrow> (\<exists>p. p \<in># P' \<and> fst (f p q))"
        unfolding mul_ext_def by auto
      {
        assume "b \<in># multiset_of qs"
        with in_multiset_of_idx[OF this] obtain i where
          "i < length qs"
          by fast
        hence i_bd: "i < length ps" using pslqs by simp
        have s1: "size {#a#} = 1" by simp
        { fix xs
          have "size (multiset_of xs) = length xs" by (induct xs) simp_all }
        note [simp] = this
        from nth_mem_multiset_of[OF i_bd] have "multiset_of ps \<noteq> {#}" by auto
        with nonempty_has_size have "0 < length ps" by simp
        moreover from P'_ne [unfolded nonempty_has_size] have "0 < size P'" .
        ultimately have "size (multiset_of ps + P') \<ge> 2"
          unfolding size_union by (simp del: length_greater_0_conv add: eval_nat_numeral)
        with P_part and s1 have "False" by simp
      }
      moreover
      {
        assume "b \<in># Q'"
        from s_P'_Q'[rule_format, OF this] obtain p where
          p_in_P': "p \<in># P'" and
          fpb: "fst (f p b)"
          by auto
        from p_in_P' and P_part and in_single_inversion[of a p]
        have "p = a" by simp
        with nfst_fab and fpb have "False" by simp
      }
      ultimately have "False" using Q_part and multi_member_last[of b] by auto
    }
    hence nfst_mul: "\<not> fst (mul_ext f [a] [b])" by fast
    show ?thesis
    proof (cases ns)
      case True
      hence snd_fab: "snd (f a b)" using Pair by simp
      hence "\<forall>i. i < length [b] \<longrightarrow> ([a] ! i, [b] ! i) \<in> ?ns" by simp
      from all_ns_ns_mul_ext[OF _ this]
      have "snd (mul_ext f [a] [b])" unfolding mul_ext_def by simp
      with nfst_fab and nfst_mul and snd_fab
        and pair_collapse[of "f a b"] and pair_collapse[of "mul_ext f [a] [b]"]
      show ?thesis by simp
    next
      case False
      hence nsnd_fab: "\<not> snd (f a b)" using Pair by simp
      {
        assume "snd (mul_ext f [a] [b])"
        then obtain ps qs P' Q' where
          P_part: "{#a#} = multiset_of ps + P'" and
          Q_part: "{#b#} = multiset_of qs + Q'" and
          pslqs: "length ps = length qs" and
          ns_ps_qs: "\<forall>i. i < length qs \<longrightarrow> (ps ! i, qs ! i) \<in> ?ns" and
          s_P'_Q': "\<forall>q. q \<in># Q' \<longrightarrow> (\<exists>p. p \<in># P' \<and> fst (f p q))"
          unfolding mul_ext_def by auto
        {
          assume "b \<in># multiset_of qs"
          with in_multiset_of_idx[OF this] obtain i where
            i_bd: "i < length qs" and
            eq_b_qsi: "b = qs ! i"
            by fast
          hence "i < length ps" using pslqs by simp
          from nth_mem_multiset_of[OF this] and P_part and in_single_inversion[of a]
          have "ps ! i = a" by simp
          with ns_ps_qs[rule_format, OF i_bd] and eq_b_qsi and nsnd_fab
          have False by simp
        }
        moreover
        {
          assume "b \<in># Q'"
          from s_P'_Q'[rule_format, OF this] obtain p where
            p_in_P': "p \<in># P'" and
            fpb: "fst (f p b)"
            by auto
          from p_in_P' and P_part and in_single_inversion[of a p]
          have "p = a" by simp
          with nfst_fab and fpb have "False" by simp
        }
        ultimately have "False" using Q_part and multi_member_last[of b] by auto
      }
      hence nsnd_mul: "\<not> snd (mul_ext f [a] [b])" by fast
      from nfst_fab and nsnd_fab and nfst_mul and nsnd_mul
       and pair_collapse[of "f a b"] and pair_collapse[of "mul_ext f [a] [b]"]
      show ?thesis by simp
    qed
  qed
qed

declare ns_mul_ext_def[simp del]
declare s_mul_ext_def[simp del]

subsection {* multisets as order on lists *}


interpretation mul_ext_list: list_order_extension 
  "\<lambda> s ns. {(as,bs). (multiset_of as, multiset_of bs) \<in> s_mul_ext ns s}"
  "\<lambda> s ns. {(as,bs). (multiset_of as, multiset_of bs) \<in> ns_mul_ext ns s}"
proof -
  let ?m = "multiset_of :: ('a list \<Rightarrow> 'a multiset)"
  let ?S = "\<lambda> s ns. {(as,bs). (?m as, ?m bs) \<in> s_mul_ext ns s}"
  let ?NS = "\<lambda> s ns. {(as,bs). (?m as, ?m bs) \<in> ns_mul_ext ns s}"
  show "list_order_extension ?S ?NS"
  proof (rule list_order_extension.intro)
    fix s ns 
    let ?s = "?S s ns"
    let ?ns = "?NS s ns"
    assume "SN_order_pair s ns"
    then interpret SN_order_pair s ns .
    interpret SN_order_pair "(s_mul_ext ns s)" "(ns_mul_ext ns s)"
      by (rule mul_ext_SN_order_pair)
    show "SN_order_pair ?s ?ns"
    proof 
      show "refl ?ns" using refl_NS unfolding refl_on_def by blast
    next
      show "?ns O ?s \<subseteq> ?s" using compat_NS_S by blast
    next
      show "?s O ?ns \<subseteq> ?s" using compat_S_NS by blast
    next
      show "trans ?ns" using trans_NS unfolding trans_def by blast
    next
      show "trans ?s" using trans_S unfolding trans_def by blast
    next
      show "SN ?s" using SN_inv_image[OF SN, of ?m, unfolded inv_image_def] .
    qed
  next
    fix S f NS as bs
    assume ass: "\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a, f b) \<in> S"
       "\<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a, f b) \<in> NS"
       "(as,bs) \<in> ?S S NS"
    show "(map f as, map f bs) \<in> ?S S NS"
      by (rule, unfold split multiset_of_map, rule s_mul_ext_map, insert ass, auto)
  next
    fix S f NS as bs
    assume ass: "\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a, f b) \<in> S"
       "\<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a, f b) \<in> NS"
       "(as,bs) \<in> ?NS S NS"    
    show "(map f as, map f bs) \<in> ?NS S NS"
      by (rule, unfold split multiset_of_map, rule ns_mul_ext_map, insert ass, auto)
  next
    fix as bs :: "'a list" and NS S :: "'a rel"
    assume ass: "length as = length bs"
    "\<And> i. i < length bs \<Longrightarrow> (as ! i, bs ! i) \<in> NS"
    show "(as,bs) \<in> ?NS S NS"
      by (rule, unfold split, rule all_ns_ns_mul_ext, insert ass, auto)
  qed
qed

end
